from keras_cv_attention_models.swin_transformer_v2.swin_transformer_v2 import (
    ExpLogitScale,
    MlpPairwisePositionalEmbedding,
    WindowAttentionMask,
    window_mhsa_with_pair_wise_positional_embedding,
    shifted_window_attention,
    SwinTransformerV2,
    SwinTransformerV2Tiny_window8,
    SwinTransformerV2Tiny_window16,
    SwinTransformerV2Small_window8,
    SwinTransformerV2Small_window16,
    SwinTransformerV2Base_window8,
    SwinTransformerV2Base_window12,
    SwinTransformerV2Base_window16,
    SwinTransformerV2Base_window24,
    SwinTransformerV2Large_window12,
    SwinTransformerV2Large_window16,
    SwinTransformerV2Large_window24,
    switch_to_deploy,
)
from keras_cv_attention_models.swin_transformer_v2.swin_transformer_v2_timm import SwinTransformerV2Tiny_ns, SwinTransformerV2Small_ns

__head_doc__ = """
Keras implementation of [Github microsoft/Swin-Transformer](https://github.com/microsoft/Swin-Transformer).
Paper [PDF 2111.09883 Swin Transformer V2: Scaling Up Capacity and Resolution](https://arxiv.org/pdf/2111.09883.pdf).
"""

__tail_doc__ = """  window_size: window_size for all blocks.
  pos_scale: If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size.
  stem_patch_size: stem patch size for stem kernel_size and strides.
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: value in {pretrained}.
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

SwinTransformerV2.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  num_heads: num heads for each stack.
  embed_dim: basic hidden dims, expand * 2 for each stack.
  use_stack_norm: boolean value if apply an additional layer_norm after each stack.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained=[None, "imagenet", "imagenet22k", "imagenet21k"]) + """
Model architectures:
  | Model                                | Params | FLOPs  | Input | Top1 Acc |
  | ------------------------------------ | ------ | ------ | ----- | -------- |
  | SwinTransformerV2Tiny_ns             | 28.3M  | 4.69G  | 224   | 81.8     |
  | SwinTransformerV2Small_ns            | 49.7M  | 9.12G  | 224   | 83.5     |
  |                                      |        |        |       |          |
  | SwinTransformerV2Tiny_window8        | 28.3M  | 5.99G  | 256   | 81.8     |
  | SwinTransformerV2Tiny_window16       | 28.3M  | 6.75G  | 256   | 82.8     |
  | SwinTransformerV2Small_window8       | 49.7M  | 11.63G | 256   | 83.7     |
  | SwinTransformerV2Small_window16      | 49.7M  | 12.93G | 256   | 84.1     |
  | SwinTransformerV2Base_window8        | 87.9M  | 20.44G | 256   | 84.2     |
  | SwinTransformerV2Base_window16       | 87.9M  | 22.17G | 256   | 84.6     |
  | SwinTransformerV2Base_window16, 22k  | 87.9M  | 22.17G | 256   | 86.2     |
  | SwinTransformerV2Base_window24, 22k  | 87.9M  | 55.89G | 384   | 87.1     |
  | SwinTransformerV2Large_window16, 22k | 196.7M | 48.03G | 256   | 86.9     |
  | SwinTransformerV2Large_window24, 22k | 196.7M | 117.1G | 384   | 87.6     |
"""

__default_doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

SwinTransformerV2Tiny_window8.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])
SwinTransformerV2Tiny_window16.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])
SwinTransformerV2Small_window8.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])
SwinTransformerV2Small_window16.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])
SwinTransformerV2Base_window8.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])
SwinTransformerV2Base_window12.__doc__ = __default_doc__.format(pretrained=[None, "imagenet21k"])
SwinTransformerV2Base_window16.__doc__ = __default_doc__.format(pretrained=[None, "imagenet", "imagenet22k"])
SwinTransformerV2Base_window24.__doc__ = __default_doc__.format(pretrained=[None, "imagenet", "imagenet22k"])
SwinTransformerV2Large_window12.__doc__ = __default_doc__.format(pretrained=[None, "imagenet21k"])
SwinTransformerV2Large_window16.__doc__ = __default_doc__.format(pretrained=[None, "imagenet22k"])
SwinTransformerV2Large_window24.__doc__ = __default_doc__.format(pretrained=[None, "imagenet22k"])
SwinTransformerV2Tiny_ns.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])
SwinTransformerV2Small_ns.__doc__ = __default_doc__.format(pretrained=[None, "imagenet"])

ExpLogitScale.__doc__ = __head_doc__ + """
Apply `inputs / tf.maximum(scale, min_value)` on given axis.

Args:
  axis: list or int number, specific axis apply scaling. Set `axis=None` for scaler `weight_shape=()`.
  init_value: weight init value. Actual using is `tf.math.log(init_value)`.
  max_value: limit scaled max value.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ExpLogitScale()
>>> print(f"{aa(tf.ones([1, 32, 32, 192])).shape = }")
# aa(tf.ones([1, 32, 32, 192])).shape = TensorShape([1, 32, 32, 192])
>>> print({ii.name:ii.shape for ii in aa.weights})
# {'divide_scale/weight:0': TensorShape([192])}

>>> bb = attention_layers.ExpLogitScale(axis=[1, 2])
>>> print(f"{bb(tf.ones([1, 32, 32, 192])).shape = }")
# bb(tf.ones([1, 32, 32, 192])).shape = TensorShape([1, 32, 32, 192])
>>> print({ii.name:ii.shape for ii in bb.weights})
# {'divide_scale_1/weight:0': TensorShape([1, 32, 32, 1])}
"""

MlpPairwisePositionalEmbedding.__doc__ = __head_doc__ + """
MLP Pair Wise Relative Positional Embedding layer.

use_absolute_pos=True input: `[batch, height, width, channel]` or `[batch, height * width, channel]`.
use_absolute_pos=True output: input + `[1, height, width, channel]` or `[1, height * width, channel]`.
use_absolute_pos=False input: `[batch, num_heads, height * width, height * width]`.
use_absolute_pos=False output: input + `[1, num_heads, height * width, height * width]`.

Args:
  hidden_dim: hidden dimension for innner MLP dense output.
  attn_height: input attention height controling initialized coords height. Default `-1` for using
      `inputs.shape[1]` if `use_absolute_pos=True and len(inputs.shape) == 4` else `sqrt(inputs.shape[-2])`.
  attn_width: input attention height controling initialized coords height. Default `-1` for using
      `inputs.shape[2]` if `use_absolute_pos=True and len(inputs.shape) == 4` else `inputs.shape[-2] // attn_height`.
  pos_scale: If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size.
      Default -1 for using `[height, width]`.
  use_absolute_pos: boolena value if using absolute or relative positional embedding.
      - `False` for using coordinates in shape `[2 * height - 1, 2 * width - 1]` and index in shape `[height * width, height * width]`.
      - `True` for using coordinates in shape `[height, width, 2]` if `len(inputs.sahpe) == 4` else `[height * width, 2]`.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.MlpPairwisePositionalEmbedding()
>>> print(f"{aa(tf.ones([9, 4, 49, 49])).shape = }")
# aa(tf.ones([9, 4, 49, 49])).shape = TensorShape([9, 4, 49, 49])
>>> print({ii.name: ii.shape for ii in aa.weights})
# {'mlp_pairwise_positional_embedding_2/hidden_weight:0': TensorShape([2, 512]),
#  'mlp_pairwise_positional_embedding_2/hidden_bias:0': TensorShape([512]),
#  'mlp_pairwise_positional_embedding_2/out:0': TensorShape([512, 4])}
>>> print(f"{aa.coords.shape = }, {aa.relative_position_index.shape = }")
# aa.coords.shape = TensorShape([169, 2]), aa.relative_position_index.shape = TensorShape([49, 49])
"""

WindowAttentionMask.__doc__ = __head_doc__ + """
Window Attention Mask layer.
No weight, just need to wrapper a layer, or will meet some error in model saving or loading.

query_blocks = `window_height * window_width`, blocks = `(height // window_height) * (width // window_width)`
input: `[batch_size * blocks, num_heads, query_blocks, query_blocks]`.
attn_mask: `[1, blocks, 1, query_blocks, query_blocks]`.
output: `[batch_size * blocks, num_heads, query_blocks, query_blocks]` + attn_mask.

attn_mask is assignd value with split blocks:
  hh_split = `[0, height - window_height, height - shift_height, height]`
  ww_split = `[0, width - window_width, width - shift_width, width]`

Args:
  height: mask height.
  width: mask width.
  window_height: mask window_height.
  window_width: mask window_width.
  shift_height: mask shift_height, should be smaller than window_height.
  shift_width: mask shift_width, should be smaller than window_width.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> height, width, window_height, window_width, shift_height, shift_width = 36, 48, 6, 6, 3, 3
>>> num_heads, query_blocks = 8, window_height * window_width
>>> aa = attention_layers.WindowAttentionMask(height, width, window_height, window_width, shift_height, shift_width)
>>> inputs = tf.ones([1 * (height // window_height) * (width // window_width), num_heads, query_blocks, query_blocks])
>>> print(f"{inputs.shape = }, {aa(inputs).shape = }")
# inputs.shape = TensorShape([48, 8, 36, 36]), aa(inputs).shape = TensorShape([48, 8, 36, 36])
>>> print(f"{aa.attn_mask.shape = }")
# aa.attn_mask.shape = TensorShape([1, 48, 1, 36, 36])
"""

window_mhsa_with_pair_wise_positional_embedding.__doc__ = __head_doc__ + """
Multi head self attention block with PairWiseRelativePositionalEmbedding, also supports mask.
Generating `attention_scores` by calculating cosine similarity between `query` and `key`,
and applying `PairWiseRelativePositionalEmbedding`.

input: `[batch * patch_height * patch_width, window_height, window_width, input_channel]`.
output: `[batch * patch_height * patch_width, window_height, window_width, num_heads * key_dim]`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  meta_hidden_dim: hidden channel for `mlp_block` applied to `PairWiseRelativePositionalEmbedding` output.
  mask: `WindowAttentionMask` layer instance. Default `None` for not applying mask.
  pos_scale: If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size.
      Default -1 for using `[height, width]` from input_shape.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  qv_bias: Boolean, if use bias for `query` and `value`.
  attn_dropout: Dropout probability for attention scores.
  out_dropout: Dropout probability for attention output.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = tf.reshape(inputs, [-1, 7, 2, 8, 2, 256])
>>> nn = tf.transpose(nn, [0, 2, 4, 1, 3, 5])
>>> nn = tf.reshape(nn, [-1, 7, 8, 256])
>>> nn = attention_layers.window_mhsa_with_pair_wise_positional_embedding(nn, num_heads=4, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 7, 8, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_qkv/kernel:0': TensorShape([256, 768]),
#  'attn_qkv/bias:0': TensorShape([768]),
#  'attn_meta_Dense_0/kernel:0': TensorShape([2, 384]),
#  'attn_meta_Dense_0/bias:0': TensorShape([384]),
#  'attn_meta_Dense_1/kernel:0': TensorShape([384, 4]),
#  'attn_meta_Dense_1/bias:0': TensorShape([4]),
#  'attn_scale/weight:0': TensorShape([1, 4, 1, 1]),
#  'attn_output/kernel:0': TensorShape([256, 256]),
#  'attn_output/bias:0': TensorShape([256])}
"""

shifted_window_attention.__doc__ = __head_doc__ + """
Window multi head self attention. Defined as function, not layer.
`window_mhsa_with_pair_wise_positional_embedding` with `window_partition` process ahead and `window_reverse` process after.
Also supports window shift.

Args:
  inputs: input tensor.
  window_size: window partition size.
  num_heads: Number of attention heads.
  shift_size: window shift retio in `(0, 1)`.
  pos_scale: If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size.
      Default -1 for using `[height, width]` from input_shape.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.shifted_window_attention(inputs, 7, num_heads=4, shift_size=0.5, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_qkv/kernel:0': TensorShape([256, 768]),
#  'attn_qkv/bias:0': TensorShape([768]),
#  'attn_meta_Dense_0/kernel:0': TensorShape([2, 384]),
#  'attn_meta_Dense_0/bias:0': TensorShape([384]),
#  'attn_meta_Dense_1/kernel:0': TensorShape([384, 4]),
#  'attn_meta_Dense_1/bias:0': TensorShape([4]),
#  'attn_scale/weight:0': TensorShape([1, 4, 1, 1]),
#  'attn_output/kernel:0': TensorShape([256, 256]),
#  'attn_output/bias:0': TensorShape([256])}
"""



from keras_cv_attention_models.edgenext.edgenext import (
    EdgeNeXt,
    EdgeNeXt_XX_Small,
    EdgeNeXt_X_Small,
    EdgeNeXt_Small,
    EdgeNeXt_Base,
    PositionalEncodingFourier,
    cross_covariance_attention,
)

__head_doc__ = """
Keras implementation of [Github mmaaz60/EdgeNeXt](https://github.com/mmaaz60/EdgeNeXt).
Paper [PDF 2206.10589 EdgeNeXt: Efficiently Amalgamated CNN-Transformer Architecture for Mobile Vision Applications](https://arxiv.org/pdf/2206.10589.pdf).
"""

__tail_doc__ = """  mlp_ratio: expand ratio for mlp blocks hidden channel.
  stem_patch_size: stem patch size for stem strides.
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: None or "imagenet" or one of ["imagenet", "usi"] for `EdgeNeXt_Small`.

Returns:
    A `keras.Model` instance.
"""

EdgeNeXt.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  num_heads: num heads for each stack.
  num_stda_layers: number of `split_depthwise_transpose_attention` blocks used in each stack tail.
  stda_split: number of inputs split for `sdta` blocks in each stack.
  stda_use_pos_emb: list of boolean value if use `PositionalEncodingFourier` positional embedding for `sdta` blocks in each stack.
  conv_kernel_size: list of int numbers, kernel_size for `conv_encoder` blocks in each stack.
  stem_width: output dimension for stem block, default `-1` for using out_channels[0]
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model             | Params | FLOPs  | Input | Top1 Acc |
  | ----------------- | ------ | ------ | ----- | -------- |
  | EdgeNeXt_XX_Small | 1.33M  | 266M   | 256   | 71.23    |
  | EdgeNeXt_X_Small  | 2.34M  | 547M   | 256   | 74.96    |
  | EdgeNeXt_Small    | 5.59M  | 1.27G  | 256   | 79.41    |
  | - usi             | 5.59M  | 1.27G  | 256   | 81.07    |
"""

EdgeNeXt_XX_Small.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

EdgeNeXt_X_Small.__doc__ = EdgeNeXt_XX_Small.__doc__
EdgeNeXt_Small.__doc__ = EdgeNeXt_XX_Small.__doc__
EdgeNeXt_Base.__doc__ = EdgeNeXt_XX_Small.__doc__

PositionalEncodingFourier.__doc__ = __head_doc__ + """
Positional Encoding Fourier layer.
Layer weight shape depends on parameter `filters` and input channel dimension only.

Args:
  filters: int number, encoded positional dimension for each point.
  temperature: temperature.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.PositionalEncodingFourier(filters=16)
>>> print(f"{aa(tf.ones([1, 12, 27, 64])).shape = }")
# aa(tf.ones([1, 12, 27, 64])).shape = TensorShape([1, 12, 27, 64])

>>> print({ii.name: ii.shape for ii in aa.weights})
# {'positional_encoding_fourier_1/ww:0': TensorShape([32, 64]), 'positional_encoding_fourier_1/bb:0': TensorShape([64])}
"""

cross_covariance_attention.__doc__ = __head_doc__ + """
Cross Covariance Attention block. Defined as function, not layer.
It's different from traditional MHSA. This is using `attention_scores` shape `[batch, num_heads, key_dim, key_dim]`,
while traditional MHSA `attention_scores` shape `[batch, num_heads, hh * ww, hh * ww]`.
Also using cosine distance between `query` and `key` calculating `attention_scores`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query, key and value. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  qkv_bias: Boolean, whether the qkv dense layer use bias vectors/matrices.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  attn_dropout: Dropout probability for attention.
  out_dropout: Dropout probability for output.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.cross_covariance_attention(inputs, num_heads=4, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_qkv/kernel:0': TensorShape([256, 768]),
#  'attn_qkv/bias:0': TensorShape([768]),
#  'attn_temperature/no_weight_decay/weight:0': TensorShape([4, 1, 1]),
#  'attn_output/kernel:0': TensorShape([256, 256]),
#  'attn_output/bias:0': TensorShape([256])}
"""



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import activation_by_name, CausalMask
from keras_cv_attention_models.gpt2.gpt2 import RunPrediction


PRETRAINED_DICT = {
    "llama2_110m": {"tiny_stories": "92b530d129f5b3c70c190669cb48b5bb"},
    "llama2_15m": {"tiny_stories": "b16894b3e40703d941e08b83767dd826"},
    "llama2_42m": {"tiny_stories": "35d924a79d6f8b1e3dcb4af8a9ae6863"},
}


@backend.register_keras_serializable(package="kecam/llama2")
class PositionalEncodingFourierRot1D(layers.Layer):
    def __init__(self, max_block_size, temperature=1e4, **kwargs):
        super().__init__(**kwargs)
        self.temperature, self.max_block_size = float(temperature), max_block_size

    def build(self, input_shape):
        # input: `[batch, ..., attn_height * attn_width, num_heads, channels // num_heads // 2, 2]`.
        # print(input_shape)
        self.channels = input_shape[-2] * input_shape[-1]
        pos_filters = self.channels // 2
        dim_t = self.temperature ** (np.arange(pos_filters, dtype="float32") / pos_filters)  # (filters,)
        grid = np.expand_dims(np.arange(self.max_block_size, dtype="float32"), -1) / dim_t
        pos_sin, pos_cos = np.expand_dims(np.sin(grid), -2), np.expand_dims(np.cos(grid), -2)
        # print(f"{pos_sin.shape = }, {pos_cos.shape = }, {height = }, {width = }, {self.channels = }")

        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("pos_sin", functional.convert_to_tensor(pos_sin, dtype=self.compute_dtype), persistent=False)
            self.register_buffer("pos_cos", functional.convert_to_tensor(pos_cos, dtype=self.compute_dtype), persistent=False)
        else:
            self.pos_sin = functional.convert_to_tensor(pos_sin, dtype=self.compute_dtype)
            self.pos_cos = functional.convert_to_tensor(pos_cos, dtype=self.compute_dtype)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        left, right = functional.unstack(inputs, axis=-1)
        pos_cos, pos_sin = self.pos_cos[: left.shape[-3]], self.pos_sin[: left.shape[-3]]
        out = functional.stack([left * pos_cos - right * pos_sin, right * pos_cos + left * pos_sin], axis=-1)
        return out

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"temperature": self.temperature, "max_block_size": self.max_block_size})
        return base_config


@backend.register_keras_serializable(package="kecam/llama2")
class RMSNorm(layers.Layer):
    def __init__(self, epsilon=1e-5, **kwargs):
        super().__init__(**kwargs)
        self.epsilon = epsilon

    def build(self, input_shape):
        self.gamma = self.add_weight(name="gamma", shape=(input_shape[-1],), initializer="ones", trainable=True)
        super().build(input_shape)

    def call(self, inputs):
        norm = inputs * functional.rsqrt(functional.reduce_mean(inputs**2, keepdims=True, axis=-1) + self.epsilon)
        return norm * self.gamma

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"epsilon": self.epsilon})
        return base_config


def apply_positional_encoding_rotary(inputs, pos_emb_layer):
    """Reshape is separated out from PositionalEncodingFourierRot1D for setting as dynamic"""
    num_heads = inputs.shape[-2]
    nn = layers.Reshape([-1, num_heads, inputs.shape[-1] // 2, 2])(inputs)
    nn = pos_emb_layer(nn)
    out = layers.Reshape([-1, num_heads, inputs.shape[-1]])(nn)
    return out


def causal_self_attention(inputs, block_size, num_heads, use_bias=False, dropout=0, name=""):
    input_channels = inputs.shape[-1]
    key_dim = input_channels // num_heads
    qq_scale = 1.0 / (float(key_dim) ** 0.5)

    query = layers.Dense(input_channels, use_bias=use_bias, name=name + "q_proj")(inputs)
    key = layers.Dense(input_channels, use_bias=use_bias, name=name + "k_proj")(inputs)
    value = layers.Dense(input_channels, use_bias=use_bias, name=name + "v_proj")(inputs)

    # Create a new one every time, as there's no weights for this layer
    rope = PositionalEncodingFourierRot1D(max_block_size=block_size, name=name + "rope")
    query = layers.Reshape([-1, num_heads, key_dim])(query)
    query = apply_positional_encoding_rotary(query, rope)
    query = functional.transpose(query, [0, 2, 1, 3])

    key = layers.Reshape([-1, num_heads, key_dim])(key)
    key = apply_positional_encoding_rotary(key, rope)
    key = functional.transpose(key, [0, 2, 3, 1])

    value = functional.transpose(layers.Reshape([-1, num_heads, key_dim])(value), [0, 2, 1, 3])

    attn = (query @ key) * qq_scale
    attn = CausalMask(block_size=block_size)(attn)
    attn = layers.Softmax(axis=-1, name=name + "attention_scores")(attn)
    attn_out = attn @ value

    output = functional.transpose(attn_out, [0, 2, 1, 3])
    output = layers.Reshape([-1, input_channels])(output)
    output = layers.Dense(input_channels, use_bias=use_bias, name=name + "o_proj")(output)
    output = layers.Dropout(dropout)(output)
    return output


def attention_fft_block(inputs, block_size, num_heads, hidden_divisible=32, use_bias=False, dropout=0, activation="swish", name=""):
    input_channels = inputs.shape[-1]
    attn = RMSNorm(name=name + "input_layernorm")(inputs)
    attn = causal_self_attention(attn, block_size, num_heads, use_bias, dropout, name=name + "self_attn.")
    attn_out = inputs + attn

    hidden_dim = 2 * 4 * input_channels // 3
    hidden_dim = hidden_divisible * ((hidden_dim + hidden_divisible - 1) // hidden_divisible)
    # print(f"{input_channels = }, {hidden_divisible = }, {hidden_dim = }")

    fft = RMSNorm(name=name + "post_attention_layernorm")(attn_out)
    fft_1 = layers.Dense(hidden_dim, use_bias=use_bias, name=name + "mlp.gate_proj")(fft)
    fft_1 = activation_by_name(fft_1, activation=activation, name=name + "mlp.gate_proj.")
    fft_3 = layers.Dense(hidden_dim, use_bias=use_bias, name=name + "mlp.up_proj")(fft)
    fft = fft_1 * fft_3
    fft = layers.Dense(input_channels, use_bias=use_bias, name=name + "mlp.down_proj")(fft)
    fft = layers.Dropout(dropout)(fft)

    return layers.Add(name=name + "output")([attn_out, fft])


def LLaMA2(
    num_blocks=12,
    embedding_size=768,
    hidden_divisible=32,
    num_heads=12,
    block_use_bias=False,
    vocab_size=50304,
    max_block_size=2048,
    include_top=True,
    dropout=0.0,
    activation="swish",
    pretrained=None,
    model_name="llama2",
    kwargs=None,
):
    inputs = layers.Input([None], dtype="int64")
    tok_emb = layers.Embedding(vocab_size, embedding_size, name="embed_tokens")(inputs)
    nn = layers.Dropout(dropout)(tok_emb)

    for block_id in range(num_blocks):
        nn = attention_fft_block(nn, max_block_size, num_heads, hidden_divisible, block_use_bias, dropout, activation, name="blocks.{}.".format(block_id))
    nn = RMSNorm(name="norm")(nn)

    if include_top:
        nn = layers.Dense(vocab_size, use_bias=False, dtype="float32", name="lm_head")(nn)

    model = models.Model(inputs, nn, name=model_name)
    model.max_block_size = max_block_size  # or model.get_layer('pos_idx').block_size
    model.run_prediction = RunPrediction(model, tokenizer="SentencePieceTokenizer")
    reload_model_weights(model, PRETRAINED_DICT, "llama2", pretrained)
    return model


@register_model
def LLaMA2_15M(max_block_size=256, vocab_size=32000, include_top=True, activation="swish", pretrained="tiny_stories", **kwargs):
    num_blocks = 6
    embedding_size = 288
    num_heads = 6
    return LLaMA2(**locals(), **kwargs, model_name="llama2_15m")


@register_model
def LLaMA2_42M(max_block_size=1024, vocab_size=32000, include_top=True, activation="swish", pretrained="tiny_stories", **kwargs):
    num_blocks = 8
    embedding_size = 512
    num_heads = 8
    return LLaMA2(**locals(), **kwargs, model_name="llama2_42m")


@register_model
def LLaMA2_110M(max_block_size=1024, vocab_size=32000, include_top=True, activation="swish", pretrained="tiny_stories", **kwargs):
    num_blocks = 12
    embedding_size = 768
    num_heads = 12
    return LLaMA2(**locals(), **kwargs, model_name="llama2_110m")


@register_model
def LLaMA2_7B(max_block_size=2048, vocab_size=32000, include_top=True, activation="swish", pretrained=None, **kwargs):
    num_blocks = 32
    embedding_size = 4096
    hidden_divisible = 256
    num_heads = 32
    return LLaMA2(**locals(), **kwargs, model_name="llama2_7b")


""" Convert pytorch weights to h5 """


def convert_huggingface_weights_to_h5(source_pt_path, save_path="AUTO", name_convert_map={}, to_fp16=False):
    from keras_cv_attention_models.download_and_load import convert_torch_weights_to_h5

    skip_weights = [".num_batches_tracked", ".inv_freq"]
    name_convert_funcs = [
        lambda name: name.replace("layers.", "blocks."),
        lambda name: name[len("model.") :] if name.startswith("model.") else name,
    ]
    weight_convert_funcs = [lambda target_name, weight: weight.T if len(weight.shape) == 2 and "embed_tokens" not in target_name else weight]  # Dense weight
    return convert_torch_weights_to_h5(source_pt_path, save_path, skip_weights, name_convert_funcs, name_convert_map, weight_convert_funcs, to_fp16)


if __name__ == "__test__":
    from keras_cv_attention_models import llama2

    mm = llama2.LLaMA2_42M()
    mm.run_prediction("As evening fell, a maiden stood at the edge of a wood. In her hands,", num_samples=3)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.download_and_load import reload_model_weights


PRETRAINED_DICT = {
    "gpt2_base": {"webtext": "0426da02ebc343d8523f5c80bb5f2eab"},
    "gpt2_large": {"webtext": "fcbca87c1d32ff4cdd70c8b9f47c7824"},
    "gpt2_medium": {"webtext": "cd2284a909af2894617df5fb6f8b91e5"},
    "gpt2_xlarge": {"webtext": ["18f8e5c068a09fdde70e5716e1b62e3f", "c8c0c2e39732d9f8e482e860cb6f8058"]},
}


@backend.register_keras_serializable(package="kecam/gpt2")
class PositionalIndex(layers.Layer):
    def __init__(self, block_size=1024, **kwargs):
        super().__init__(**kwargs)
        self.block_size = block_size
        self.use_layer_as_module = True

    def build(self, input_shape):
        pos_idx = np.arange(0, self.block_size, dtype="int64")[None]
        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("pos_idx", functional.convert_to_tensor(pos_idx, dtype="int64"), persistent=False)
        else:
            self.pos_idx = functional.convert_to_tensor(pos_idx, dtype="int64")
        super().build(input_shape)

    def call(self, inputs):
        # print(inputs.shape)
        return self.pos_idx[:, : inputs.shape[-1]]

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"block_size": self.block_size})
        return base_config


@backend.register_keras_serializable(package="kecam/gpt2")
class CausalMask(layers.Layer):
    def __init__(self, block_size, **kwargs):
        super().__init__(**kwargs)
        self.block_size = block_size
        self.use_layer_as_module = True

    def build(self, input_shape):
        causal_mask = (1 - np.tri(self.block_size).astype("float32")[None, None]) * -65504  # instead of -1e10, fit float16
        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("causal_mask", functional.convert_to_tensor(causal_mask, dtype=self.compute_dtype), persistent=False)
        else:
            self.causal_mask = functional.convert_to_tensor(causal_mask, dtype=self.compute_dtype)
        super().build(input_shape)

    def call(self, inputs):
        return inputs + self.causal_mask[:, :, : inputs.shape[2], : inputs.shape[3]]

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"block_size": self.block_size})
        return base_config


def causal_self_attention(inputs, block_size, num_heads, use_bias, dropout, name=""):
    input_channels = inputs.shape[-1]
    key_dim = input_channels // num_heads
    qq_scale = 1.0 / (float(key_dim) ** 0.5)

    qkv = layers.Dense(3 * input_channels, use_bias=use_bias, name=name + "qkv")(inputs)
    query, key, value = functional.split(qkv, 3, axis=-1)
    query = functional.transpose(layers.Reshape([-1, num_heads, key_dim])(query), [0, 2, 1, 3])
    key = functional.transpose(layers.Reshape([-1, num_heads, key_dim])(key), [0, 2, 3, 1])
    value = functional.transpose(layers.Reshape([-1, num_heads, key_dim])(value), [0, 2, 1, 3])

    attn = (query @ key) * qq_scale
    attn = CausalMask(block_size=block_size)(attn)
    attn = layers.Softmax(axis=-1, name=name + "attention_scores")(attn)
    attn_out = attn @ value

    output = functional.transpose(attn_out, [0, 2, 1, 3])
    output = layers.Reshape([-1, input_channels])(output)
    output = layers.Dense(input_channels, use_bias=use_bias, name=name + "attn_out")(output)
    output = layers.Dropout(dropout)(output)
    return output


def attention_mlp_block(inputs, block_size, num_heads, use_bias, dropout, activation="gelu/app", name=""):
    input_channels = inputs.shape[-1]
    attn = layers.LayerNormalization(axis=-1, name=name + "attn_ln")(inputs)
    attn = causal_self_attention(attn, block_size, num_heads, use_bias, dropout, name=name + "attn.")
    attn_out = inputs + attn

    mlp = layers.LayerNormalization(axis=-1, name=name + "mlp_ln")(attn_out)
    mlp = layers.Dense(4 * input_channels, use_bias=use_bias, name=name + "mlp.0")(mlp)
    mlp = functional.gelu(mlp, approximate=True) if activation == "gelu/app" else layers.Activation(activation, name=name)(mlp)
    mlp = layers.Dense(input_channels, use_bias=use_bias, name=name + "mlp.2")(mlp)
    mlp = layers.Dropout(dropout)(mlp)

    return layers.Add(name=name + "output")([attn_out, mlp])


def GPT2(
    num_blocks=12,
    embedding_size=768,
    num_heads=12,
    block_use_bias=True,
    vocab_size=50304,
    max_block_size=1024,
    include_top=True,
    dropout=0.0,
    activation="gelu/app",
    pretrained=None,
    model_name="gpt2",
    kwargs=None,
):
    inputs = layers.Input([None], dtype="int64")
    pos_idx = PositionalIndex(block_size=max_block_size, name="pos_idx")(inputs)

    tok_emb = layers.Embedding(vocab_size, embedding_size, name="wte")(inputs)
    pos_emb = layers.Embedding(max_block_size, embedding_size, name="wpe")(pos_idx)
    nn = layers.Dropout(dropout)(tok_emb + pos_emb)

    for block_id in range(num_blocks):
        nn = attention_mlp_block(nn, max_block_size, num_heads, block_use_bias, dropout, activation=activation, name="blocks.{}.".format(block_id))
    nn = layers.LayerNormalization(axis=-1, name="ln_f")(nn)

    if include_top:
        nn = layers.Dense(vocab_size, use_bias=False, dtype="float32", name="lm_head")(nn)

    model = models.Model(inputs, nn, name=model_name)
    model.max_block_size = max_block_size  # or model.get_layer('pos_idx').block_size
    model.run_prediction = RunPrediction(model)
    if pretrained == "huggingface":
        load_weights_from_huggingface(model, save_path="~/.keras/models")
    else:
        reload_model_weights(model, PRETRAINED_DICT, "gpt2", pretrained)
    return model


@register_model
def GPT2_Base(max_block_size=1024, vocab_size=50257, include_top=True, activation="gelu/app", pretrained="webtext", **kwargs):
    return GPT2(**locals(), **kwargs, model_name="gpt2_base")


@register_model
def GPT2_Medium(max_block_size=1024, vocab_size=50257, include_top=True, activation="gelu/app", pretrained="webtext", **kwargs):
    num_blocks = 24
    embedding_size = 1024
    num_heads = 16
    return GPT2(**locals(), **kwargs, model_name="gpt2_medium")


@register_model
def GPT2_Large(max_block_size=1024, vocab_size=50257, include_top=True, activation="gelu/app", pretrained="webtext", **kwargs):
    num_blocks = 36
    embedding_size = 1280
    num_heads = 20
    return GPT2(**locals(), **kwargs, model_name="gpt2_large")


@register_model
def GPT2_XLarge(max_block_size=1024, vocab_size=50257, include_top=True, activation="gelu/app", pretrained="webtext", **kwargs):
    num_blocks = 48
    embedding_size = 1600
    num_heads = 25
    return GPT2(**locals(), **kwargs, model_name="gpt2_xlarge")


""" Load weights and run prediction functions """


class RunPrediction:
    def __init__(self, model, tokenizer="GPT2Tokenizer"):
        self.model, self.tokenizer, self.built = model, tokenizer, False

    def build(self):
        from keras_cv_attention_models.clip import tokenizer

        self.tokenizer = getattr(tokenizer, self.tokenizer)() if isinstance(self.tokenizer, str) else self.tokenizer
        self.max_block_size = self.model.max_block_size
        self.vocab_size = self.model.output_shape[-1]
        self.vocab_indexes = np.arange(self.vocab_size)
        self.built = True

    @staticmethod
    def softmax_numpy(inputs, axis=-1):
        exp_inputs = np.exp(inputs - np.max(inputs, axis=axis))
        return exp_inputs / np.sum(exp_inputs, keepdims=True, axis=axis)

    def __call__(self, inputs, num_samples=1, max_new_tokens=500, temperature=0.8, top_k=200, stop_at_eot=True):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape (b,t)) and complete
        the sequence max_new_tokens times, feeding the predictions back into the model each time.
        Most likely you'll want to make sure to be in model.eval() mode of operation for this.

        Args:
          num_samples: number of samples to draw.
          max_new_tokens: number of tokens generated in each sample.
          temperature: 1.0 = no change, < 1.0 = less random, > 1.0 = more random, in predictions.
          top_k: retain only the top_k most likely tokens, clamp others to have 0 probability.
          stop_at_eot: boolean value if stop at eot_token output.
        """
        if not self.built:
            self.build()
        if backend.is_torch_backend:
            device = next(self.model.parameters()).device

        start_ids = np.array(self.tokenizer.encode(inputs, add_sot=True))
        for k in range(num_samples):
            inputs_idxes = start_ids
            print(self.tokenizer.decode(inputs_idxes.tolist()), end="", flush=True)
            for _ in range(max_new_tokens):
                # if the sequence context is growing too long we must crop it at block_size
                idx_cond = inputs_idxes if inputs_idxes.shape[-1] <= self.max_block_size else inputs_idxes[-self.max_block_size :]
                # forward the model to get the logits for the index in the sequence
                cur_inputs = functional.convert_to_tensor(idx_cond, dtype="int64")[None]
                if backend.is_torch_backend:
                    cur_inputs = cur_inputs.to(device)
                logits = self.model(cur_inputs)
                # pluck the logits at the final step and scale by desired temperature
                logits = logits[:, -1, :] / temperature
                logits = logits.detach().cpu() if hasattr(logits, "detach") else logits
                logits = logits.numpy() if hasattr(logits, "numpy") else np.array(logits)  # np.array is for JAX

                if top_k == 0 or top_k == 1:
                    pick = np.argmax(logits, axis=-1)
                else:
                    if top_k > 0:
                        # optionally crop the logits to only the top k options
                        threshold_pos = min(top_k, self.vocab_size)
                        logits_threshold = np.sort(logits)[:, -threshold_pos]
                        logits[logits < logits_threshold[:, None]] = -float("Inf")
                    # sample from the distribution
                    probs = self.softmax_numpy(logits.astype("float32"), axis=-1)
                    pick = np.array([np.random.choice(self.vocab_indexes, p=prob) for prob in probs])
                inputs_idxes = np.concatenate([inputs_idxes, pick], axis=-1)
                next_word = self.tokenizer.decode(inputs_idxes[-1:].tolist())
                print(next_word, end="", flush=True)

                if stop_at_eot and pick == self.tokenizer.eot_token or pick == self.tokenizer.sot_token:
                    break
            print("\n---------------")


def load_weights_from_huggingface(model, save_name=None, save_path=".", force=False):
    import os

    model_type_map = {"gpt2_base": "gpt2", "gpt2_medium": "gpt2-medium", "gpt2_large": "gpt2-large", "gpt2_xlarge": "gpt2-xl"}
    if model.name not in model_type_map:
        print("No pretrained available, model will be randomly initialized.")
        return

    pretrained = "huggingface"
    save_name = save_name if save_name is not None else "{}_{}.h5".format(model.name, pretrained)
    save_path = os.path.join(os.path.expanduser(save_path), save_name)
    if not force and os.path.exists(save_path):
        print("Load previously saved model:", save_path)
        model.load_weights(save_path)
        return
    else:
        print("Convert and load weights from huggingface")

    from transformers import GPT2LMHeadModel

    model_type = model_type_map[model.name]
    source_state_dict = GPT2LMHeadModel.from_pretrained(model_type).state_dict()

    """ state_dict_stack_by_layer """
    stacked_state_dict = {}
    for kk, vv in source_state_dict.items():
        if kk.endswith(".attn.bias") or kk.endswith(".attn.masked_bias") or kk.endswith(".num_batches_tracked"):
            continue

        split_kk = kk.split(".")
        vv = vv.numpy() if hasattr(vv, "numpy") else vv

        # split_kk[-1] in ["weight", "bias", "running_mean", "running_var", "gain"]
        layer_name = ".".join(split_kk[:-1])
        stacked_state_dict.setdefault(layer_name, []).append(vv)
    stacked_state_dict["lm_head"] = [ii.T for ii in stacked_state_dict["lm_head"]]

    """ keras_reload_stacked_state_dict """
    target_names = [ii.name for ii in model.layers if len(ii.weights) != 0]
    for target_name, source_name in zip(target_names, stacked_state_dict.keys()):
        print(">>>> Load {} weights from {}".format(target_name, source_name))
        target_layer = model.get_layer(target_name)
        source_weights = stacked_state_dict[source_name]
        print("    Source: {}, Target: {}".format([ii.shape for ii in source_weights], [ii.shape for ii in target_layer.get_weights()]))

        if hasattr(target_layer, "set_weights_channels_last"):
            target_layer.set_weights_channels_last(source_weights)  # Kecam PyTorch backend
        else:
            target_layer.set_weights(source_weights)

    print(">>>> Save to:", save_path)
    if hasattr(model, "save"):
        model.save(save_path)
    else:
        model.save_weights(save_path)  # Kecam PyTorch backend


if __name__ == "__test__":
    from keras_cv_attention_models import gpt2

    mm = gpt2.GPT2_Base()
    mm.run_prediction("hello world")



import torch
import math
import torch.nn.functional as F
from functools import partial
from keras_cv_attention_models.pytorch_backend.layers import (
    _ZeroPadding,
    _ReshapeDynamic,
    _ResizeDynamic,
    _GatherND,
    Add,
    Concatenate,
    Equal,
    GraphNode,
    Lambda,
    Shape,
    compute_conv_output_size,
)


def wrapper(func, inputs, output_shape=None, name=None):
    if isinstance(inputs, GraphNode) or (isinstance(inputs, list) and isinstance(inputs[0], GraphNode)):
        return Lambda(func, output_shape=output_shape, name=name)(inputs)
    else:
        return func(inputs)


def abs(inputs, name=None):
    return wrapper(torch.abs, inputs, name=name)


def argmax(inputs, axis=None, output_type="int64", name=None):
    axis = 0 if axis is None else (len(inputs.shape) + axis if axis < 0 else axis)
    output_shape = [ii for id, ii in enumerate(inputs.shape) if id != axis]
    return wrapper(partial(torch.argmax, dim=axis), inputs, output_shape=output_shape, name=name)


def argsort(inputs, axis=-1, direction="ASCENDING", stable=False, name=None):
    descending = direction.upper() == "DESCENDING"
    return wrapper(partial(torch.argsort, dim=axis, descending=descending, stable=stable), inputs, output_shape=inputs.shape, name=name)


def assign(parameter, data):
    parameter.data = torch.tensor(data, dtype=parameter.dtype)


def cast(inputs, dtype="float32"):
    if isinstance(inputs, torch.Tensor) and (inputs.dtype == dtype or str(inputs.dtype).endswith(str(dtype))):
        return inputs
    elif dtype == torch.float32 or dtype == "float32":
        return inputs.float()
    elif dtype == torch.float16 or dtype == "float16" or dtype == "half":
        return inputs.half()
    else:
        dtype = dtype if isinstance(dtype, str) else str(dtype).split(".")[-1]
        return torch.tensor(inputs, dtype=getattr(torch, dtype))


def clip_by_value(inputs, clip_value_min, clip_value_max, name=None):
    return wrapper(partial(torch.clip, min=clip_value_min, max=clip_value_max), inputs, name=name)


def concat(inputs, axis, name=None):
    return Concatenate(axis=axis, name=name)(inputs)


def convert_to_tensor(inputs, dtype="float32"):
    return torch.tensor(inputs, dtype=getattr(torch, dtype) if isinstance(dtype, str) else dtype)


def cos(inputs, name=None):
    return wrapper(torch.cos, inputs, name=name)


def embedding_lookup(params, ids, max_norm=None, name=None):
    """
    >>> import math, torch
    >>> inputs = torch.randint(0, 1000, size=[32])
    >>> params = torch.rand([1000, 320])
    >>> print(np.allclose(torch.functional.F.embedding(inputs, params), tf.nn.embedding_lookup(params.numpy(), inputs.numpy())))
    # True
    """
    # return wrapper(partial(F.embedding, max_norm=max_norm)), [params, ids], name=name)
    return F.embedding(ids, params, max_norm=max_norm)


def equal(x, y, name=None):
    return Equal(name=name)([x, y]) if isinstance(y, GraphNode) else wrapper(lambda inputs: torch.eq(inputs, y), x, output_shape=x.shape, name=name)


def exp(inputs, name=None):
    return wrapper(torch.exp, inputs, name=name)


def expand_dims(inputs, axis, name=None):
    return wrapper(partial(torch.unsqueeze, dim=axis), inputs, name=name)


def extract_patches(inputs, sizes=1, strides=1, rates=1, padding="valid", data_format="channels_last", compressed=True, name=None):
    """
    tf.image.extract_patches -> out shape [batch, patch_height, patch_width, kernel_size * kernel_szie * channel]
    torch.functional.F.unfold -> out shape [batch, channel * kernel_size * kernel_szie, patch_height * patch_width]
    - compressed=False -> out shape [batch, patch_height, patch_width, kernel_size, kernel_szie, channel]

    >>> import torch
    >>> from keras_cv_attention_models.pytorch_backend import functional
    >>> kernel_size = 3
    >>> image = np.zeros([256, 256, 3])
    >>> image[:, :, 1] += 128 # G
    >>> image[:, :, 2] += 256 # B
    >>> aa = np.expand_dims(image.astype("float32"), 0)
    >>> tf_out = tf.image.extract_patches(aa, sizes=[1, kernel_size, kernel_size, 1], strides=[1, 2, 2, 1], rates=[1, 1, 1, 1], padding='valid')
    >>> torch_out = functional.extract_patches(torch.from_numpy(image[None].astype('float32')), kernel_size, 2)
    >>> print(f"{np.allclose(tf_out, torch_out.detach()) = }")
    """
    channel = inputs.shape[1] if data_format == "channels_first" else inputs.shape[-1]
    height, width = inputs.shape[2:] if data_format == "channels_first" else inputs.shape[1:-1]

    kernel_size = sizes[1] if isinstance(sizes, (list, tuple)) else sizes
    strides = strides[1] if isinstance(strides, (list, tuple)) else strides
    # dilation_rate can be 2 different values, used in DiNAT
    dilation_rate = (rates if len(rates) == 2 else rates[1:3]) if isinstance(rates, (list, tuple)) else (rates, rates)

    padding = padding.lower()
    pad_value = kernel_size // 2 if padding == "same" else 0
    unfold = torch.nn.Unfold(kernel_size=kernel_size, dilation=dilation_rate, padding=pad_value, stride=strides)
    patch_size = compute_conv_output_size((height, width), kernel_size, strides, padding, dilation_rate=dilation_rate)
    if data_format == "channels_first":
        if compressed:
            target_shape = [-1, channel * kernel_size * kernel_size, *patch_size]
        else:
            target_shape = [-1, channel, kernel_size, kernel_size, *patch_size]
        module = lambda xx: unfold(xx).contiguous().view(target_shape)
    else:
        pre_perm_shape = [-1, channel, kernel_size * kernel_size, patch_size[0] * patch_size[1]]
        perm = [0, 3, 2, 1]
        if compressed:
            target_shape = [-1, *patch_size, kernel_size * kernel_size * channel]
        else:
            target_shape = [-1, *patch_size, kernel_size, kernel_size, channel]
        module = lambda xx: unfold(xx.permute([0, 3, 1, 2])).contiguous().view(pre_perm_shape).permute(perm).contiguous().view(target_shape)
    return wrapper(module, inputs, name=name)


def gather(inputs, indices, axis=None, batch_dims=0, name=None):
    """Defaults axis=None means the first non-batch dimension"""
    axis = batch_dims if axis is None else (len(inputs.shape) + axis if axis < 0 else axis)
    indices = tuple([slice(None)] * axis + [indices])
    return inputs[indices]  # .contiguous()


def gather_nd(inputs, indices, batch_dims=0, name=None):
    # """Defaults axis=None means the first non-batch dimension"""
    # return inputs[indices.tolist()]  # .contiguous()
    # print(indices.shape[0])
    if isinstance(inputs, GraphNode) or isinstance(indices, GraphNode):
        return _GatherND(batch_dims=batch_dims, name=name)([inputs, indices])
    else:
        return inputs[indices.T.tolist()] if len(indices.shape) > 1 else inputs[tuple(indices.tolist())]  # .contiguous()


def gelu(inputs, approximate=False, name=None):
    return wrapper(partial(F.gelu, approximate="tanh" if approximate else "none"), inputs, name=name)


def l2_normalize(inputs, axis=None, epsilon=1e-12, name=None):
    return wrapper(partial(F.normalize, p=2.0, dim=axis, eps=epsilon), inputs, name=name)


def linspace(start, stop, num, name=None, axis=0):
    # return Lambda(partial(torch.linspace, start=start, end=stop, steps=num), name=name)(inputs)
    return torch.linspace(start=start, end=stop, steps=num)


def log(inputs, name=None):
    return wrapper(partial(torch.log), inputs, name=name)


def matmul(xx, yy, transpose_a=False, transpose_b=False, name=None):
    return wrapper(lambda inputs: torch.matmul(inputs[0].T if transpose_a else inputs[0], inputs[1].T if transpose_b else inputs[1]), [xx, yy], name=name)


def moments(inputs, axes, shift=None, keepdims=False, name=None):
    return wrapper(partial(torch.var_mean, dim=axes, keepdim=keepdims), inputs, name=name)


def maximum(xx, yy, name=None):
    if isinstance(yy, torch.Tensor):
        return wrapper(lambda inputs: torch.maximum(inputs[0], inputs[1]), [xx, yy], name=name)
    else:  # maximum doesn't support scalar value
        return wrapper(lambda inputs: torch.clip(xx, min=yy), [xx, yy], name=name)


def minimum(xx, yy, name=None):
    if isinstance(yy, torch.Tensor):
        return wrapper(lambda inputs: torch.minimum(inputs[0], inputs[1]), [xx, yy], name=name)
    else:  # maximum doesn't support scalar value
        return wrapper(lambda inputs: torch.clip(xx, max=yy), [xx, yy], name=name)


def non_max_suppression_with_scores(boxes, scores, max_output_size=100, iou_threshold=0.5, score_threshold=-math.inf, soft_nms_sigma=0.0, name=None):
    from torchvision.ops import nms, batched_nms

    valid_scores_index = torch.where(scores > score_threshold)[0]
    nms_index = nms(boxes=boxes[valid_scores_index], scores=scores[valid_scores_index], iou_threshold=iou_threshold)
    actual_index = valid_scores_index[nms_index]
    return actual_index, scores[actual_index]
    # from tensorflow import image
    #
    # if hasattr(boxes, "detach"):
    #     boxes = boxes.detach().numpy()
    # if hasattr(scores, "detach"):
    #     scores = scores.detach().numpy()
    # rr, nms_scores = image.non_max_suppression_with_scores(boxes, scores, max_output_size, iou_threshold, score_threshold, soft_nms_sigma)
    # return torch.from_numpy(rr.numpy().astype("int64")), torch.from_numpy(nms_scores.numpy().astype("float32"))


def norm(inputs, ord="euclidean", axis=1, keepdims=False, name=None):
    return wrapper(partial(torch.norm, p=2, dim=axis, keepdim=keepdims), inputs, name=name)


def pad(inputs, paddings, mode="CONSTANT", constant_values=0, name=None):
    """
    torch pad is like `[left, right, top, bottom]`
    >>> aa = tf.pad(np.array([[[1, 2, 3], [4, 5, 6]]]), [[0, 0], [1, 2], [3, 4]])
    >>> bb = torch.functional.F.pad(torch.tensor([[[1, 2, 3], [4, 5, 6]]]), [3, 4, 1, 2, 0, 0])
    >>> np.allclose(aa, bb.detach())
    """
    # F.pad doesn't support 0 shape inputs, throws error while compute_output_shape
    # pad = []
    # for pp in paddings[::-1]:
    #     pad += pp
    # return Lambda(partial(F.pad, pad=pad, mode=mode.lower(), value=constant_values), name=name)(inputs)
    return _ZeroPadding(padding=paddings, mode=mode.lower(), value=constant_values)(inputs)


def pow(inputs, exponent, name=None):
    return wrapper(partial(torch.pow, exponent=exponent), inputs, name=name)


def range(start, limit=None, delta=1, dtype=None, name="range"):
    if limit is None:
        start, limit = 0, start
    dtype = dtype and (getattr(torch, dtype) if isinstance(dtype, str) else dtype)
    if isinstance(limit, Shape) and (None in limit or -1 in limit):
        return wrapper(lambda limit: torch.arange(start=start, end=limit, step=delta, dtype=dtype), limit, output_shape=[None], name=name)
    else:
        return torch.arange(start=start, end=limit, step=delta, dtype=dtype)


def reduce_max(inputs, axis=None, keepdims=False, name=None):
    # return wrapper(partial(torch.max, dim=axis, keepdim=keepdims), inputs, name=name)
    return wrapper(lambda inputs: torch.max(inputs, dim=axis, keepdim=keepdims)[0], inputs, name=name)


def reduce_mean(inputs, axis=None, keepdims=False, name=None):
    return wrapper(partial(torch.mean, dim=axis, keepdim=keepdims), inputs, name=name)


def reduce_sum(inputs, axis=None, keepdims=False, name=None):
    axis = () if axis is None else axis
    if isinstance(inputs, (list, tuple)) and axis == 0:
        return Add(name=name)(inputs)
    else:
        # return wrapper(lambda xx: xx.sum(dim=axis, keepdim=keepdims), inputs, name=name)
        return wrapper(partial(torch.sum, dim=axis, keepdim=keepdims), inputs, name=name)


def relu(inputs, name=None):
    return wrapper(F.relu, inputs, name=name)


def relu6(inputs, name=None):
    return wrapper(F.relu6, inputs, name=name)


def repeat(inputs, repeats, axis, name=None):
    """
    >>> aa = np.arange(6).reshape(2, 3)
    >>> torch_out = torch.expand_copy(torch.unsqueeze(torch.from_numpy(aa), 1), [2, 2, 3]).reshape(4, 3)
    >>> tf_out = tf.repeat(aa, repeats=2, axis=0)
    >>> np.allclose(torch_out.detach(), tf_out)
    """
    if inputs.shape[axis] is None:
        expand_shape = [1] * len(inputs.shape)
        expand_shape.insert(axis + 1, repeats)
        out_shape = [(-1 if ii is None or ii == -1 else ii * repeats) if dim == axis else ii for dim, ii in enumerate(inputs.shape)]
        # return wrapper(lambda inputs: torch.expand_copy(torch.unsqueeze(inputs, axis + 1), expand_shape).contiguous().view(out_shape), inputs, name=name)
        return wrapper(lambda xx: xx.unsqueeze(axis + 1).repeat(expand_shape).contiguous().view(out_shape), inputs, name=name)
    else:
        return wrapper(partial(torch.repeat_interleave, repeats=repeats, dim=axis), inputs, name=name)
    # expand_shape.insert(axis + 1, repeats)
    # expand_shape = tuple(expand_shape)
    # out_shape = [(-1 if ii is None or ii == -1 else ii * repeats) if dim == axis else ii for dim, ii in enumerate(inputs.shape)]
    # return wrapper(lambda inputs: torch.reshape(torch.expand_copy(torch.unsqueeze(inputs, axis + 1), expand_shape), out_shape), inputs, name=name)


def reshape(inputs, shape, name=None):
    if isinstance(inputs, GraphNode) and (isinstance(shape, Shape) and (None in shape[1:] or -1 in shape[1:])):  # If dynamic
        return _ReshapeDynamic(target_shape=list(shape), name=name)([inputs, shape])
    else:
        # return wrapper(partial(torch.reshape, shape=shape), inputs, name=name)
        return wrapper(lambda inputs: inputs.contiguous().view(shape), inputs, name=name)


def resize(inputs, size, method="bilinear", preserve_aspect_ratio=False, antialias=False, name=None):
    if isinstance(inputs, GraphNode) and (isinstance(size, Shape) and (None in size or -1 in size)):  # If dynamic
        return _ResizeDynamic(method=method, preserve_aspect_ratio=preserve_aspect_ratio, antialias=antialias, name=name)([inputs, size])
    elif isinstance(inputs, GraphNode):
        return Lambda(partial(F.interpolate, size=list(size), mode=method, antialias=antialias), name=name)(inputs)  # [TODO] align_corners
        # return Resize(size=size, mode=method, preserve_aspect_ratio=preserve_aspect_ratio, antialias=antialias, name=name)(inputs)  # [TODO] align_corners
    else:  # called directly
        inputs = inputs if isinstance(inputs, torch.Tensor) else torch.tensor(inputs)
        if len(inputs.shape) == 3:
            return F.interpolate(inputs[None], size=list(size), mode=method, antialias=antialias)[0]
        else:
            return F.interpolate(inputs, size=list(size), mode=method, antialias=antialias)


def rsqrt(inputs, name=None):
    return wrapper(torch.rsqrt, inputs, name=name)


def shape(inputs):
    return Shape(inputs) if isinstance(inputs, GraphNode) else inputs.shape
    # return inputs.shape


def sigmoid(inputs, axis=None, name=None):
    return wrapper(torch.sigmoid, inputs, name=name)


def sign(inputs, name=None):
    return wrapper(torch.sign, inputs, name=name)


def sin(inputs, name=None):
    return wrapper(torch.sin, inputs, name=name)


def softmax(inputs, axis=None, name=None):
    return wrapper(partial(torch.softmax, dim=-1 if axis is None else axis), inputs, name=name)


def softplus(inputs, name=None):
    return wrapper(F.softplus, inputs, name=name)


def split(inputs, num_or_size_splits, axis=0, num=None, name=None):
    axis = (len(inputs.shape) + axis) if axis < 0 else axis
    split_axis_shape = inputs.shape[axis]
    assert split_axis_shape is not None

    if isinstance(num_or_size_splits, int):
        # split_axis_shape, num_or_size_splits = 5, 3 -> size_splits [2, 2, 1]
        split_size = math.ceil(split_axis_shape / num_or_size_splits)
        size_splits = [split_size] * num_or_size_splits  # doesn't matter if the last one exceeding split_axis_shape
    else:
        size_splits = num_or_size_splits
        size_splits = [0 if ii is None or ii == -1 else ii for ii in size_splits]
        num_unknown_dim = sum([ii == 0 for ii in size_splits])
        assert num_unknown_dim < 2, "At most one unknown dimension in num_or_size_splits: {}".format(num_or_size_splits)

        if num_unknown_dim == 1:
            size_splits = [(split_axis_shape - sum(size_splits)) if ii == 0 else ii for ii in size_splits]
    # [112, 112] -> [slice(0, 112), slice(112, 224)]
    split_slices = [slice(int(sum(size_splits[:id])), sum(size_splits[: id + 1])) for id, ii in enumerate(size_splits)]

    pre_axis_slice = [slice(None)] * axis
    return [inputs[tuple([*pre_axis_slice, split_slice])] for split_slice in split_slices]


def sqrt(inputs, name=None):
    return wrapper(torch.sqrt, inputs, name=name)


def square(inputs, name=None):
    return wrapper(torch.square, inputs, name=name)


def squeeze(inputs, axis=None, name=None):
    return wrapper(partial(torch.squeeze, dim=axis), inputs, name=name)


def stack(inputs, axis=0, name=None):
    return wrapper(partial(torch.stack, dim=axis), inputs, name=name)


def tanh(inputs, name=None):
    return wrapper(F.tanh, inputs, name=name)


def tile(inputs, multiples, name=None):
    return wrapper(lambda xx: xx.repeat(multiples), inputs, name=name)


def top_k(inputs, k=1, sorted=True, name=None):
    return wrapper(partial(torch.topk, k=k, sorted=sorted), inputs, name=name)


def transpose(inputs, perm=None, conjugate=False, name=None):
    return wrapper(lambda xx: xx.T, inputs, name=name) if perm is None else wrapper(partial(torch.permute, dims=perm), inputs, name=name)


def unstack(inputs, axis=0, name=None):
    axis = len(inputs.shape) + axis if axis < 0 else axis
    axis_shape = inputs.shape[axis]
    assert axis_shape is not None

    pre_axis_slice = [slice(None)] * axis
    return [inputs[tuple([*pre_axis_slice, index])] for index in range(axis_shape)]



from keras_cv_attention_models import *
_sub_modules = {__name__ + "." + kk: vv for kk, vv in locals().items() if not kk.startswith("_")}

import sys as _sys
_sys.modules.update(_sub_modules)

from keras_cv_attention_models.version import __version__



from keras_cv_attention_models.fasternet.fasternet import FasterNet, FasterNetT0, FasterNetT1, FasterNetT2, FasterNetS, FasterNetM, FasterNetL

__head_doc__ = """
Keras implementation of [Github JierunChen/FasterNet](https://github.com/JierunChen/FasterNet).
Paper [PDF 2303.03667 Run, Don’t Walk: Chasing Higher FLOPS for Faster Neural Networks ](https://arxiv.org/pdf/2303.03667.pdf).
"""

__tail_doc__ = """  window_ratios: window split ratio. Each stack will calculate `window_size = (height // window_ratio, width // window_ratio)` .
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: one of None or "imagenet".
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

FasterNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  embed_dim: basic hidden dims, expand * 2 for each stack.
  patch_size: int value for stem kernel size and strides.
  mlp_ratio: expand ratio for mlp blocks hidden channel.
  partial_conv_ratio: float value for partial channles appling `Conv2D` in each block.
  output_conv_filter: int value for filters of `Conv2D` block before output block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model       | Params | FLOPs  | Input | Top1 Acc |
  | ----------- | ------ | ------ | ----- | -------- |
  | FasterNetT0 | 3.9M   | 0.34G  | 224   | 71.9     |
  | FasterNetT1 | 7.6M   | 0.85G  | 224   | 76.2     |
  | FasterNetT2 | 15.0M  | 1.90G  | 224   | 78.9     |
  | FasterNetS  | 31.1M  | 4.55G  | 224   | 81.3     |
  | FasterNetM  | 53.5M  | 8.72G  | 224   | 83.0     |
  | FasterNetL  | 93.4M  | 15.49G | 224   | 83.5     |
"""

FasterNetT0.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

FasterNetT1.__doc__ = FasterNetT0.__doc__
FasterNetT2.__doc__ = FasterNetT0.__doc__
FasterNetS.__doc__ = FasterNetT0.__doc__
FasterNetM.__doc__ = FasterNetT0.__doc__
FasterNetL.__doc__ = FasterNetT0.__doc__



import torch

epsilon = 1e-7


class Loss:
    def __init__(self, reduction="AUTO", name=None):
        self.reduction, self.name = reduction, name

    def __call__(self, y_true, y_pred, sample_weight=None):
        pass


def categorical_crossentropy(y_true, y_pred, from_logits=False, label_smoothing=0.0, axis=-1):
    """
    # from_logits=False
    >>> import torch, tensorflow as tf
    >>> from keras_cv_attention_models.pytorch_backend import losses
    >>> xx, yy = tf.random.uniform([24, 10]), tf.one_hot(tf.random.uniform([24], 0, 10, dtype='int32'), 10)
    >>> tf_out = tf.losses.categorical_crossentropy(yy, xx, from_logits=False).numpy().mean()
    >>> torch_out = losses.categorical_crossentropy(torch.from_numpy(yy.numpy()), torch.from_numpy(xx.numpy()), from_logits=False)
    >>> print(tf_out, torch_out, np.allclose(tf_out, torch_out))
    >>> # 2.681877 tensor(2.6819) True
    # from_logits=True
    >>> tf_out = tf.losses.categorical_crossentropy(yy, xx, from_logits=True).numpy().mean()
    >>> torch_out = losses.categorical_crossentropy(torch.from_numpy(yy.numpy()), torch.from_numpy(xx.numpy()), from_logits=True)
    >>> print(tf_out, torch_out, np.allclose(tf_out, torch_out))
    >>> # 2.3364408 tensor(2.3364) True
    """
    if from_logits:
        return torch.nn.functional.cross_entropy(y_pred, y_true.argmax(-1), label_smoothing=label_smoothing)
    else:
        y_pred = y_pred / y_pred.sum(dim=axis, keepdim=True)
        y_pred = y_pred.clamp_(epsilon, 1.0 - epsilon)
        return -(y_true * y_pred.log()).sum(dim=axis).mean()


def sparse_categorical_crossentropy(y_true, y_pred, from_logits=False, label_smoothing=0.0, axis=-1):
    """
    # from_logits=False
    >>> import torch, tensorflow as tf
    >>> from keras_cv_attention_models.pytorch_backend import losses
    >>> xx, yy = tf.random.uniform([24, 10]), tf.random.uniform([24], 0, 10, dtype='int64')
    >>> tf_out = tf.losses.sparse_categorical_crossentropy(yy, xx, from_logits=False).numpy().mean()
    >>> torch_out = losses.sparse_categorical_crossentropy(torch.from_numpy(yy.numpy()), torch.from_numpy(xx.numpy()), from_logits=False)
    >>> print(tf_out, torch_out, np.allclose(tf_out, torch_out))
    >>> # 2.677911 tensor(2.6779) True
    # from_logits=True
    >>> tf_out = tf.losses.sparse_categorical_crossentropy(yy, xx, from_logits=True).numpy().mean()
    >>> torch_out = losses.sparse_categorical_crossentropy(torch.from_numpy(yy.numpy()), torch.from_numpy(xx.numpy()), from_logits=True)
    >>> print(tf_out, torch_out, np.allclose(tf_out, torch_out))
    >>> # 2.3503969 tensor(2.3504) True
    """
    if from_logits:
        return torch.nn.functional.cross_entropy(y_pred, y_true, label_smoothing=label_smoothing)
    else:
        y_pred = y_pred / y_pred.sum(dim=axis, keepdim=True)
        y_pred = y_pred.clamp_(epsilon, 1.0 - epsilon)
        y_true = torch.nn.functional.one_hot(y_true, y_pred.shape[-1])
        return -(y_true * y_pred.log()).sum(dim=axis).mean()


class MeanSquaredError(Loss):
    """
    >>> from keras_cv_attention_models.pytorch_backend import losses
    >>> aa = np.random.uniform(size=[4, 42, 42, 3]).astype("float32")
    >>> bb = np.random.uniform(size=[4, 42, 42, 3]).astype("float32")
    >>> print(f"{keras.losses.MeanSquaredError()(aa, bb).numpy() = }")
    # keras.losses.MeanSquaredError()(aa, bb).numpy() = 0.16724217
    >>> print(f"{losses.MeanSquaredError()(torch.from_numpy(aa), torch.from_numpy(bb)) = }")
    # losses.MeanSquaredError()(torch.from_numpy(aa), torch.from_numpy(bb)) = tensor(0.1672)
    """

    def __init__(self, reduction="AUTO", name="mean_squared_error"):
        super().__init__(reduction=reduction, name=name)

    def __call__(self, y_true, y_pred, sample_weight=None):
        return torch.functional.F.mse_loss(y_pred, y_true)



from keras_cv_attention_models.mobilenetv3_family.mobilenetv3 import MobileNetV3
from keras_cv_attention_models.models import register_model


def FBNetV3(
    num_blocks=[2, 4, 1, 4, 1, 4, 1, 5, 1, 5, 1],
    out_channels=[16, 24, 40, 40, 72, 72, 120, 120, 184, 184, 224],
    expands=[1, [4, 2, 2, 2], 5, 3, 5, 3, 5, 3, 6, 4, 6],
    kernel_sizes=[3, 5, 5, 5, 5, 3, 3, 5, 3, 5, 5],
    strides=[1, 2, 2, 1, 2, 1, 1, 1, 2, 1, 1],
    activations="hard_swish",
    se_ratios=[0, 0, 0.25, 0.25, 0, 0, 0.25, 0.25, 0.25, 0.25, 0.25],
    se_activation=("hard_swish", "hard_sigmoid_torch"),
    se_limit_round_down=0.95,
    use_expanded_se_ratio=False,
    output_num_features=1984,
    use_output_feature_bias=False,
    model_name="fbnetv3",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    return MobileNetV3(**locals(), **kwargs)


@register_model
def FBNetV3B(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return FBNetV3(**locals(), model_name="fbnetv3_b", **kwargs)


@register_model
def FBNetV3D(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 1, 4, 1, 4, 1, 6, 1, 5, 1]
    out_channels = [16, 24, 40, 40, 72, 72, 128, 128, 208, 208, 240]
    expands = [1, [5, 2, 2, 2, 2, 2], 4, 3, 5, 3, 5, 3, 6, 5, 6]
    kernel_sizes = [3, 3, 5, 3, 3, 3, 3, 5, 3, 5, 5]
    stem_width = 24
    return FBNetV3(**locals(), model_name="fbnetv3_d", **kwargs)


@register_model
def FBNetV3G(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 5, 1, 4, 1, 4, 1, 8, 1, 6, 2]
    out_channels = [24, 40, 56, 56, 104, 104, 160, 160, 264, 264, 288]
    expands = [1, [4, 2, 2, 2, 2], 4, 3, 5, 3, 5, 3, 6, 5, 6]
    kernel_sizes = [3, 5, 5, 5, 5, 3, 3, 5, 3, 5, 5]
    stem_width = 32
    return FBNetV3(**locals(), model_name="fbnetv3_g", **kwargs)



from keras_cv_attention_models.repvit.repvit import RepViT, RepViT_M1, RepViT_M2, RepViT_M3, switch_to_deploy


__head_doc__ = """
Keras implementation of [Github THU-MIG/RepViT](https://github.com/THU-MIG/RepViT).
Paper [PDF 2307.09283 RepViT: Revisiting Mobile CNN From ViT Perspective](https://arxiv.org/pdf/2307.09283.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  deploy: boolean value if build a fused model. **Evaluation only, not good for training**.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `hard_swish`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: top dropout rate if top layers is included. Default 0.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  use_distillation: Boolean value if output `distill_head`. Default `False`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

RepViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of block for each stack.
  out_channels: output channels for each stack.
  stem_width: channel dimension output for stem block, default -1 for using out_channels[0].
  se_ratio: float value for se_ratio for each stack, will use `se_module` every other block in each stack if > 0.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                   | Params | FLOPs | Input | Top1 Acc |
  | ----------------------- | ------ | ----- | ----- | -------- |
  | RepViT_M1, distillation | 5.10M  | 0.82G | 224   | 78.5     |
  | - deploy=True           | 5.07M  | 0.82G | 224   | 78.5     |
  | RepViT_M2, distillation | 8.28M  | 1.35G | 224   | 80.6     |
  | - deploy=True           | 8.25M  | 1.35G | 224   | 80.6     |
  | RepViT_M3, distillation | 10.2M  | 1.87G | 224   | 81.4     |
  | - deploy=True           | 10.12M | 1.87G | 224   | 81.4     |
"""

RepViT_M1.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

RepViT_M2.__doc__ = RepViT_M1.__doc__
RepViT_M3.__doc__ = RepViT_M1.__doc__



import tensorflow as tf
from keras_cv_attention_models.imagenet.data import init_mean_std_by_rescale_mode, tf_imread, random_crop_and_resize_image, build_custom_dataset


def image_process(image, image_size=(224, 224), is_train=True):
    image = tf_imread(image)
    if is_train:
        image = random_crop_and_resize_image(image, image_size, scale=(0.9, 1.0), method="bicubic", antialias=True)[0]
    else:
        image = tf.image.resize(image, image_size, method="bicubic", antialias=True)
    image = tf.cast(image, tf.float32)
    image.set_shape([*image_size, 3])
    return image


def init_dataset(data_path, caption_tokenizer, batch_size=64, image_size=224, rescale_mode="torch"):
    dataset, total_images, num_classes, num_channels = build_custom_dataset(data_path, with_info=True, caption_tokenizer=caption_tokenizer)

    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    image_size = image_size if isinstance(image_size, (list, tuple)) else [image_size, image_size]

    AUTOTUNE, buffer_size, seed = tf.data.AUTOTUNE, batch_size * 100, None
    train_pre_batch = lambda data_point: (image_process(data_point["image"], image_size, is_train=True), data_point["caption"])
    y_true = tf.range(batch_size)
    train_post_batch = lambda xx, caption: (((xx - mean) / std, caption), y_true)

    train_dataset = dataset["train"]
    train_dataset = train_dataset.shuffle(buffer_size, seed=seed).map(train_pre_batch, num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.batch(batch_size, drop_remainder=True).map(train_post_batch, num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.prefetch(buffer_size=AUTOTUNE)

    test_dataset = dataset.get("validation", dataset.get("test", None))
    if test_dataset is not None:
        test_pre_batch = lambda data_point: (image_process(data_point["image"], image_size, is_train=False), data_point["caption"])
        test_dataset = test_dataset.map(test_pre_batch, num_parallel_calls=AUTOTUNE)
        test_dataset = test_dataset.batch(batch_size, drop_remainder=True).map(train_post_batch)

    return train_dataset, test_dataset



import math
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    add_with_layer_scale_and_drop_block,
    ChannelAffine,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    mlp_block,
    multi_head_self_attention,
    output_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "davit_t": {"imagenet": "040a215cbcf1c0ce06db665cdcf6f9ac"},
    "davit_s": {"imagenet": "b95558071c639815f4ab2e9d09a4141f"},
    "davit_b": {"imagenet": "89e50de7a70ea7b2404f8f57369d8015"},
}


def multi_head_self_attention_channel(
    inputs, num_heads=4, key_dim=0, out_shape=None, out_weight=True, qkv_bias=False, out_bias=False, attn_dropout=0, output_dropout=0, name=None
):
    input_channel = inputs.shape[-1]
    input_blocks = inputs.shape[1:-1]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    # qk_scale = float(1.0 / tf.math.sqrt(tf.cast(key_dim, "float32")))
    qk_scale = 1.0 / (float(key_dim) ** 0.5)
    out_shape = input_channel if out_shape is None or not out_weight else out_shape
    qkv_out = num_heads * key_dim

    qkv = functional.reshape(inputs, [-1, int(np.prod(input_blocks)), inputs.shape[-1]]) if len(inputs.shape) > 3 else inputs
    qkv = layers.Dense(qkv_out * 3, use_bias=qkv_bias, name=name and name + "qkv")(qkv)
    value, query, key = functional.split(qkv, 3, axis=-1)  # Matching weights from PyTorch
    query = functional.transpose(functional.reshape(query, [-1, query.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  # [batch, num_heads, key_dim, hh * ww]
    key = functional.transpose(functional.reshape(key, [-1, key.shape[1], num_heads, key_dim]), [0, 2, 1, 3])  # [batch, num_heads, hh * ww, key_dim]
    value = functional.transpose(functional.reshape(value, [-1, value.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  #  [batch, num_heads, key_dim, hh * ww]

    # attention_scores = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([query, key]) * qk_scale  # [batch, num_heads, key_dim, key_dim]
    attention_scores = (query @ key) * qk_scale
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attention_scores)
    attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores) if attn_dropout > 0 else attention_scores

    # value = [batch, num_heads, key_dim, hh * ww], attention_output = [batch, num_heads, key_dim, hh * ww]
    # attention_output = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([attention_scores, value])
    attention_output = attention_scores @ value
    attention_output = functional.transpose(attention_output, [0, 3, 1, 2])  # [batch, hh * ww, num_heads, key_dim]
    attention_output = functional.reshape(attention_output, [-1, *input_blocks, num_heads * key_dim])
    # print(f">>>> {attention_output.shape = }, {attention_scores.shape = }")

    if out_weight:
        # [batch, hh, ww, num_heads * key_dim] * [num_heads * key_dim, out] --> [batch, hh, ww, out]
        attention_output = layers.Dense(out_shape, use_bias=out_bias, name=name and name + "output")(attention_output)
    attention_output = layers.Dropout(output_dropout, name=name and name + "out_drop")(attention_output) if output_dropout > 0 else attention_output
    return attention_output


def pad_to_divisible_by_window_size(inputs, window_size):
    window_size = window_size if isinstance(window_size, (list, tuple)) else [window_size, window_size]
    window_height = window_size[0] if window_size[0] < inputs.shape[1] else inputs.shape[1]
    window_width = window_size[1] if window_size[1] < inputs.shape[2] else inputs.shape[2]

    # window_partition, partition windows, ceil mode
    patch_height, patch_width = int(math.ceil(inputs.shape[1] / window_height)), int(math.ceil(inputs.shape[2] / window_width))
    padding_height, padding_width = patch_height * window_height - inputs.shape[1], patch_width * window_width - inputs.shape[2]
    # print(f">>>> window_attention {inputs.shape = }, {padding_height = }, {padding_width = }")
    if padding_height or padding_width:
        inputs = functional.pad(inputs, [[0, 0], [0, padding_height], [0, padding_width], [0, 0]])
    return inputs, window_height, window_width, padding_height, padding_width


def reverse_padded_for_window_size(inputs, padding_height, padding_width):
    if padding_height or padding_width:
        inputs = inputs[:, : inputs.shape[1] - padding_height, : inputs.shape[2] - padding_width, :]  # In case padding_height or padding_width is 0
    return inputs


def window_partition(inputs, window_height, window_width=-1):
    """[B, patch_height * window_height, patch_width * window_width, channel] -> [B * patch_height * patch_width, window_height, window_width, channel]"""
    input_channel = inputs.shape[-1]
    window_width = window_width if window_width > 0 else window_height
    patch_height, patch_width = inputs.shape[1] // window_height, inputs.shape[2] // window_width
    if patch_height == 1 and patch_width == 1:
        return inputs

    # print(f">>>> window_attention {inputs.shape = }, {patch_height = }, {patch_width = }, {window_height = }, {window_width = }")
    # [batch * patch_height, window_height, patch_width, window_width * channel], limit transpose perm <= 4
    nn = functional.reshape(inputs, [-1, window_height, patch_width, window_width * input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, patch_width, window_height, window_width * channel]
    nn = functional.reshape(nn, [-1, window_height, window_width, input_channel])  # [batch * patch_height * patch_width, window_height, window_width, channel]
    return nn


def window_reverse(inputs, patch_height, patch_width, window_height, window_width):
    """[B * patch_height * patch_width, window_height, window_width, channel] -> [B, patch_height * window_height, patch_width * window_width, channel]"""
    input_channel = inputs.shape[-1]
    patch_width = patch_width if patch_width > 0 else patch_height
    # window_height, window_width = inputs.shape[1], inputs.shape[2]
    if patch_height == 1 and patch_width == 1:
        return functional.reshape(inputs, [-1, patch_height * window_height, patch_width * window_width, input_channel])

    # [batch * patch_height, patch_width, window_height, window_width * input_channel], limit transpose perm <= 4
    nn = functional.reshape(inputs, [-1, patch_width, window_height, window_width * input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, window_height, patch_width, window_width * input_channel]
    nn = functional.reshape(nn, [-1, patch_height * window_height, patch_width * window_width, input_channel])
    return nn


def grid_window_partition(inputs, window_height, window_width=-1):
    """[B, window_height * patch_height, window_width * patch_width , channel] -> [B * patch_height * patch_width, window_height, window_width, channel]"""
    input_channel = inputs.shape[-1]
    window_width = window_width if window_width > 0 else window_height
    patch_height, patch_width = inputs.shape[1] // window_height, inputs.shape[2] // window_width
    if patch_height == 1 and patch_width == 1:
        return inputs

    nn = functional.reshape(inputs, [-1, window_height, patch_height, window_width * patch_width * input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch, patch_height, window_height, window_width * patch_width * input_channel]
    nn = functional.reshape(nn, [-1, window_height * window_width, patch_width, input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, patch_width, window_height * window_width, input_channel]
    nn = functional.reshape(nn, [-1, window_height, window_width, input_channel])
    return nn


def grid_window_reverse(inputs, patch_height, patch_width, window_height, window_width):
    """[B * patch_height * patch_width, window_height, window_width, channel] -> [B, window_height * patch_height, window_width * patch_width , channel]"""
    input_channel = inputs.shape[-1]
    patch_width = patch_width if patch_width > 0 else patch_height
    # window_height, window_width = inputs.shape[1], inputs.shape[2]
    if patch_height == 1 and patch_width == 1:
        return functional.reshape(inputs, [-1, window_height * patch_height, window_width * patch_width, input_channel])

    nn = functional.reshape(inputs, [-1, patch_width, window_height * window_width, input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch * patch_height, window_height * window_width, patch_width, input_channel]
    nn = functional.reshape(nn, [-1, patch_height, window_height, window_width * patch_width * input_channel])
    nn = functional.transpose(nn, [0, 2, 1, 3])  # [batch, window_height, patch_height, window_width * patch_width * input_channel]
    nn = functional.reshape(nn, [-1, window_height * patch_height, window_width * patch_width, input_channel])
    return nn


def window_attention(inputs, window_size, num_heads=4, is_grid=False, attention_block=None, data_format="channels_last", name=None, **kwargs):
    inputs = inputs if data_format == "channels_last" else functional.transpose(inputs, [0, 2, 3, 1])
    inputs, window_height, window_width, padding_height, padding_width = pad_to_divisible_by_window_size(inputs, window_size)
    patch_height, patch_width = inputs.shape[1] // window_height, inputs.shape[2] // window_width

    if is_grid:
        nn = grid_window_partition(inputs, window_height, window_width)
    else:
        nn = window_partition(inputs, window_height, window_width)

    if attention_block:
        nn = nn if data_format == "channels_last" else functional.transpose(nn, [0, 3, 1, 2])
        nn = attention_block(nn, num_heads=num_heads, name=name, **kwargs)
        nn = nn if data_format == "channels_last" else functional.transpose(nn, [0, 2, 3, 1])
    else:
        nn = multi_head_self_attention(nn, num_heads=num_heads, qkv_bias=True, out_bias=True, name=name)

    # window_reverse, merge windows
    if is_grid:
        nn = grid_window_reverse(nn, patch_height, patch_width, window_height, window_width)
    else:
        nn = window_reverse(nn, patch_height, patch_width, window_height, window_width)

    nn = reverse_padded_for_window_size(nn, padding_height, padding_width)
    return nn if data_format == "channels_last" else functional.transpose(nn, [0, 3, 1, 2])


def conv_positional_encoding(inputs, kernel_size=3, use_norm=False, activation="gelu", name=""):
    nn = depthwise_conv2d_no_bias(inputs, kernel_size, padding="same", use_bias=True, name=name)
    if use_norm:
        nn = layer_norm(nn, name=name)
    if activation is not None:
        nn = activation_by_name(nn, activation, name=name)
    return layers.Add(name=name + "output")([inputs, nn])


def davit_block(
    inputs, window_size, num_heads=4, use_channel_attn=False, mlp_ratio=4, mlp_drop_rate=0, attn_drop_rate=0, drop_rate=0, layer_scale=-1, name=None
):
    input_channel = inputs.shape[-1 if backend.image_data_format() == "channels_last" else 1]

    pre_attn = conv_positional_encoding(inputs, 3, use_norm=False, activation=None, name=name + "pre_attn_cpe_")
    attn = pre_attn if backend.image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(pre_attn)
    attn = layer_norm(attn, axis=-1, name=name + "attn_")
    if use_channel_attn:
        attn = multi_head_self_attention_channel(attn, num_heads, qkv_bias=True, out_bias=True, name=name + "channel_attn_")
    else:
        attn = window_attention(attn, window_size, num_heads, name=name + "attn_")
    attn = attn if backend.image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(attn)
    # print(f"{pre_attn.shape = }, {attn.shape = }, {inputs.shape = }")
    attn_out = add_with_layer_scale_and_drop_block(pre_attn, attn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "1_")

    pre_ffn = conv_positional_encoding(attn_out, 3, use_norm=False, activation=None, name=name + "pre_ffn_cpe_")
    mlp = pre_ffn if backend.image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(pre_ffn)
    mlp = layer_norm(mlp, axis=-1, name=name + "mlp_")
    mlp = mlp_block(mlp, int(input_channel * mlp_ratio), drop_rate=mlp_drop_rate, use_conv=False, activation="gelu", name=name + "mlp_")
    mlp = mlp if backend.image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(mlp)
    return add_with_layer_scale_and_drop_block(pre_ffn, mlp, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "2_")


def DaViT(
    num_blocks=[2, 2, 6, 2],
    out_channels=[96, 192, 384, 768],
    num_heads=[3, 6, 12, 24],
    stem_width=-1,
    stem_patch_size=4,
    # window_size=7,
    window_ratio=32,
    mlp_ratio=4,
    layer_scale=-1,
    input_shape=(224, 224, 3),
    num_classes=1000,
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="davit",
    kwargs=None,
):
    """Patch stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv2d_no_bias(inputs, stem_width, kernel_size=7, strides=stem_patch_size, use_bias=True, padding="same", name="stem_")
    nn = layer_norm(nn, name="stem_")
    # window_size = [input_shape[0] // window_ratio, input_shape[1] // window_ratio]
    window_size = [int(math.ceil(input_shape[0] / window_ratio)), int(math.ceil(input_shape[1] / window_ratio))]
    # window_size = window_size[:2] if isinstance(window_size, (list, tuple)) else [window_size, window_size]

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, num_head) in enumerate(zip(num_blocks, out_channels, num_heads)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            ds_name = stack_name + "downsample_"
            nn = layer_norm(nn, name=ds_name)
            # Set use_torch_padding=False, as kernel_size == 2, otherwise shape will be enlarged by 1
            nn = conv2d_no_bias(nn, out_channel, kernel_size=2, strides=2, use_bias=True, padding="same", use_torch_padding=False, name=ds_name)
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            use_channel_attn = False if block_id % 2 == 0 else True
            nn = davit_block(nn, window_size, num_head, use_channel_attn, mlp_ratio, drop_rate=block_drop_rate, layer_scale=layer_scale, name=block_name)
            global_block_id += 1
    nn = layer_norm(nn, name="pre_output_")

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "davit", pretrained)
    return model


@register_model
def DaViT_T(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 6, 2]
    return DaViT(**locals(), model_name="davit_t", **kwargs)


@register_model
def DaViT_S(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 18, 2]
    return DaViT(**locals(), model_name="davit_s", **kwargs)


@register_model
def DaViT_B(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 18, 2]
    out_channels = [128, 256, 512, 1024]
    num_heads = [4, 8, 16, 32]
    return DaViT(**locals(), model_name="davit_b", **kwargs)


@register_model
def DaViT_L(input_shape=(384, 384, 3), num_classes=1000, classifier_activation="softmax", pretrained=None, **kwargs):
    num_blocks = [2, 2, 18, 2]
    out_channels = [192, 384, 768, 1536]
    num_heads = [6, 12, 24, 48]
    return DaViT(**locals(), model_name="davit_l", **kwargs)


@register_model
def DaViT_H(input_shape=(512, 512, 3), num_classes=1000, classifier_activation="softmax", pretrained=None, **kwargs):
    num_blocks = [2, 2, 18, 2]
    out_channels = [256, 512, 1024, 2048]
    num_heads = [8, 16, 32, 64]
    return DaViT(**locals(), model_name="davit_h", **kwargs)


@register_model
def DaViT_G(input_shape=(512, 512, 3), num_classes=1000, classifier_activation="softmax", pretrained=None, **kwargs):
    num_blocks = [2, 2, 24, 6]
    out_channels = [384, 768, 1536, 3072]
    num_heads = [12, 24, 48, 96]
    return DaViT(**locals(), model_name="davit_g", **kwargs)



from keras_cv_attention_models.visualizing.visualize_filters import visualize_filters
from keras_cv_attention_models.visualizing.gradcam_heatmap import make_gradcam_heatmap, make_and_apply_gradcam_heatmap
from keras_cv_attention_models.visualizing.attention_score_maps import plot_attention_score_maps
from keras_cv_attention_models.plot_func import tensorboard_parallel_coordinates_plot


visualize_filters.__doc__ = """
Copied and modified from: https://keras.io/examples/vision/visualizing_what_convnets_learn/
Displaying the visual patterns that convnet filters respond to.
Will create input images that maximize the activation of specific filters in a target layer.
Such images represent a visualization of the pattern that the filter responds to.

Note: models using `Conv2D` with `groups != 1` not supporting on CPU. Needs backward steps.

Args:
  model: keras model used for visualizing.
  layer_name: target layer name in model for visualizing. Default "auto" means using the last layer.
  filter_index_list: channel indexes for visualizing.
  input_shape: model input_shape. Default `None` means using `model.input_shape`.
  rescale_mode: one of ["tf", "torch", "caffe"]. Default `auto` means using model preset.
  iterations: total steps running gradient ascent.
  optimizer: one of ["SGD", "RMSprop", "Adam"] or specified inited optimizer.
  learning_rate: learning rate in runnung gradient ascent. Default "auto" means {"SGD": 10.0, "RMSprop": 1.0, "Adam": 0.1}.
  value_range: indicates input image value range, processed to `[128 - 128 * value_range, 128 + 128 * value_range]`.
  random_magnitude: random magnitude for `tf.roll` and `random_rotation`. basic random value is `4` and `1`.
  crop_border: crop output image border pixels, calculated as:
      `hh_crop, ww_crop = input_shape[0] * crop_border, input_shape[1] * crop_border`
      `img = img[hh_crop:-hh_crop, ww_crop:-ww_crop]`
  base_size: base plotting size for a single image.

Returns:
  losses, filter_images, ax

Example:
>>> from keras_cv_attention_models import visualizing, resnest
>>> mm = resnest.ResNest50()
>>> losses, filter_images, ax = visualizing.visualize_filters(mm, "stack3_block6_out", range(10))
>>> print(f"{losses[0] = }, {filter_images[0].shape = }")
# losses[0] = 23.950336, filter_images[0].shape = (174, 174, 3)
"""

make_gradcam_heatmap.__doc__ = """
Copied and modified from: https://keras.io/examples/vision/grad_cam/
Grad-CAM class activation visualization. Obtain a class activation heatmap for an image classification model.

Args:
  model: keras model used for visualizing.
  processed_image: preprocessed image that can be directly used as model input.
  layer_name: target layer name in model for visualizing. Default "auto" means using the last layer with `len(output_shape) == 4`.
  pred_index: specified visualizing prediction index. Used for image containing multi classes.
      Default `None` means using max probability one.
  use_v2: set False for a simple version, not a big difference.

Returns:
  heatmap, preds

Example:
>>> from keras_cv_attention_models import visualizing, resnest
>>> mm = resnest.ResNest50()
>>> url = 'https://upload.wikimedia.org/wikipedia/commons/b/bc/Free%21_%283987584939%29.jpg'
>>> img = plt.imread(keras.utils.get_file('aa.jpg', url))
>>> img = tf.expand_dims(tf.image.resize(img, mm.input_shape[1:-1]), 0)
>>> img = keras.applications.imagenet_utils.preprocess_input(img, mode='torch')
>>> heatmap, preds = visualizing.make_gradcam_heatmap(mm, img, 'stack4_block3_out')
>>> print(f"{preds.shape = }, {heatmap.shape = }, {heatmap.max() = }, {heatmap.min() = }")
# preds.shape = (1, 1000), heatmap.shape = (7, 7), heatmap.max() = 1.0, heatmap.min() = 0.0
>>> print(keras.applications.imagenet_utils.decode_predictions(preds)[0][0])
# ('n02110063', 'malamute', 0.54736596)

>>> plt.imshow(heatmap)
"""

make_and_apply_gradcam_heatmap.__doc__ = """
Copied and modified from: https://keras.io/examples/vision/grad_cam/
Grad-CAM class activation visualization. Create and plot a superimposed visualization heatmap on image.

Args:
  model: keras model used for visualizing.
  image: Original image for visualizing.
  layer_name: target layer name in model for visualizing. Default "auto" means using the last layer with `len(output_shape) == 4`.
  rescale_mode: image value rescale mode. Mostly used one is "tf" or "torch". Default `auto` means using model preset.
  pred_index: specified visualizing prediction index. Used for image containing multi classes.
      Default `None` means using max probability one.
  alpha: heatmap superimposed alpha over image.
  use_v2: set False for a simple version, not a big difference.
  plot: set False to disable plot image.

Returns:
  superimposed_img, heatmap, preds

Example:
>>> from keras_cv_attention_models import visualizing, resnest
>>> mm = resnest.ResNest50()
>>> url = 'https://upload.wikimedia.org/wikipedia/commons/b/bc/Free%21_%283987584939%29.jpg'
>>> img = plt.imread(keras.utils.get_file('aa.jpg', url))
>>> superimposed_img, heatmap, preds = visualizing.make_and_apply_gradcam_heatmap(mm, img, "stack4_block3_out")
"""

plot_attention_score_maps.__doc__ = """
Visualizing model attention score maps, superimposed with specific image.

Args:
  model: keras model used for visualizing.
  image: Original image for visualizing.
  rescale_mode: image value rescale mode. Mostly used one is "tf" or "torch". Default `auto` means using model preset.
  attn_type: Specify model attention type. Currently supporting `["beit", "levit", "bot", "coatnet", "halo"]`.
      Default "auto" means decide from model name. Technically any attention scores in the same format with
      specified `attn_type` can be supported. Like using `attn_type="beit"` for `VIT` models.
  rows: Specify number of plotted rows. Default `-1` means auto adjust.
  base_size: base plotting size for a single image.

Returns:
  mask, cum_mask, fig

Example:
>>> from keras_cv_attention_models import botnet, halonet, beit, levit, visualizing
>>> url = 'https://upload.wikimedia.org/wikipedia/commons/b/bc/Free%21_%283987584939%29.jpg'
>>> imm = plt.imread(keras.utils.get_file('aa.jpg', url))
>>> _ = visualizing.plot_attention_score_maps(beit.BeitBasePatch16(), imm, rescale_mode='tf', rows=2)
"""

tensorboard_parallel_coordinates_plot.__doc__ = """
Parallel coordinates plotting using TensorBoard, wrapper of https://www.tensorflow.org/tensorboard/hyperparameter_tuning_with_hparams

Args:
  dataframe: data saved in `pd.DataFrame`.
  metrics_name: dataframe column name used as result metrics.
  metrics_display_name: speficy another name for displaying metrics. Default None for using `metrics_name`.
  skip_columns: list value of column names which will be ignored. Default `[]`.
  log_dir: tensorboard log saving directory. Default 'logs/hparam_tuning'.
      Note need to manually clean saved data if plotting new data using same log_dir.

Example:
>>> import pandas as pd
>>> from keras_cv_attention_models import visualizing
>>> aotnet50_imagnet_results = {
>>>   "optimizer": ["lamb", "lamb", "adamw", "adamw", "adamw"],
>>>   "rescale_mode": ["torch", "tf", "torch", "torch", "torch"],
>>>   "lr_base": [8e-3, 8e-3, 4e-3, 4e-3, 8e-3],
>>>   "weight_decay": [0.05, 0.05, 0.05, 0.02, 0.02],
>>>   "accuracy": [78.48, 78.31, 77.92, 78.06, 78.27],
>>> }
>>> aa = pd.DataFrame(aotnet50_imagnet_results)
>>> visualizing.tensorboard_parallel_coordinates_plot(aa, 'accuracy', log_dir="logs/aotnet50_imagnet_results")
>>> # >>>> Start tensorboard by: ! tensorboard --logdir logs/aotnet50_imagnet_results
>>> # >>>> Then select `HPARAMS` -> `PARALLEL COORDINATES VIEW`
>>> ! tensorboard --logdir logs/aotnet50_imagnet_results
>>> # TensorBoard 2.8.0 at http://localhost:6006/ (Press CTRL+C to quit)
"""



"""
Creates a EfficientNetV1 Model as defined in: EfficientNetV1: Self-training with Noisy Student improves ImageNet classification.
arXiv preprint arXiv:1911.04252.
"""
import math
from keras_cv_attention_models.efficientnet.efficientnet_v2 import EfficientNetV2
from keras_cv_attention_models.models import register_model


def get_expanded_width_depth(width, depth, fix_head_stem=False):
    out_channels = [ii * width for ii in [16, 24, 40, 80, 112, 192, 320]]
    depthes = [int(math.ceil(ii * depth)) for ii in [1, 2, 2, 3, 3, 4, 1]]
    if fix_head_stem:
        depthes[0], depthes[-1] = 1, 1
        first_conv_filter, output_conv_filter = 32, 1280
    else:
        first_conv_filter = 32 * width
        output_conv_filter = 1280 * width
    return out_channels, depthes, first_conv_filter, output_conv_filter


def EfficientNetV1(
    expands=[1, 6, 6, 6, 6, 6, 6],
    out_channels=[16, 24, 40, 80, 112, 192, 320],
    depthes=[1, 2, 2, 3, 3, 4, 1],
    strides=[1, 2, 2, 2, 1, 2, 1],
    se_ratios=[0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25],
    first_conv_filter=32,
    output_conv_filter=1280,
    kernel_sizes=[3, 3, 5, 3, 5, 5, 3],
    drop_connect_rate=0.2,
    pretrained="noisy_student",
    model_name="EfficientNetV1",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    return EfficientNetV2(**locals(), **kwargs)


@register_model
def EfficientNetV1B0(input_shape=(224, 224, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.0, 1.0)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b0", **kwargs)


@register_model
def EfficientNetV1B1(input_shape=(240, 240, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.0, 1.1)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b1", **kwargs)


@register_model
def EfficientNetV1B2(input_shape=(260, 260, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.1, 1.2)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b2", **kwargs)


@register_model
def EfficientNetV1B3(input_shape=(300, 300, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.2, 1.4)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b3", **kwargs)


@register_model
def EfficientNetV1B4(input_shape=(380, 380, 3), num_classes=1000, dropout=0.4, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.4, 1.8)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b4", **kwargs)


@register_model
def EfficientNetV1B5(input_shape=(456, 456, 3), num_classes=1000, dropout=0.4, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.6, 2.2)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b5", **kwargs)


@register_model
def EfficientNetV1B6(input_shape=(528, 528, 3), num_classes=1000, dropout=0.5, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.8, 2.6)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b6", **kwargs)


@register_model
def EfficientNetV1B7(input_shape=(600, 600, 3), num_classes=1000, dropout=0.5, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(2.0, 3.1)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-b7", **kwargs)


@register_model
def EfficientNetV1L2(input_shape=(800, 800, 3), num_classes=1000, dropout=0.5, classifier_activation="softmax", pretrained="noisy_student", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(4.3, 5.3)
    first_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter)
    output_conv_filter = kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), model_name="efficientnet_v1-l2", **kwargs)


""" EfficientNetV1Lite models from EfficientDet """


# https://github.com/google/automl/tree/master/efficientdet/backbone/efficientnet_lite_builder.py
@register_model
def EfficientNetV1Lite0(input_shape=(320, 320, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained=None, **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.0, 1.0, fix_head_stem=True)
    first_conv_filter, output_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter), kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), se_ratios=[0] * len(depthes), is_fused=False, model_name="efficientnet_v1-lite0", **kwargs)


@register_model
def EfficientNetV1Lite1(input_shape=(384, 384, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained=None, **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.0, 1.1, fix_head_stem=True)
    first_conv_filter, output_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter), kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), se_ratios=[0] * len(depthes), is_fused=False, model_name="efficientnet_v1-lite1", **kwargs)


@register_model
def EfficientNetV1Lite2(input_shape=(448, 448, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained=None, **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.1, 1.2, fix_head_stem=True)
    first_conv_filter, output_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter), kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), se_ratios=[0] * len(depthes), is_fused=False, model_name="efficientnet_v1-lite2", **kwargs)


@register_model
def EfficientNetV1Lite3(input_shape=(512, 512, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained=None, **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.2, 1.4, fix_head_stem=True)
    first_conv_filter, output_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter), kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), se_ratios=[0] * len(depthes), is_fused=False, model_name="efficientnet_v1-lite3", **kwargs)


@register_model
def EfficientNetV1Lite4(input_shape=(640, 640, 3), num_classes=1000, dropout=0.4, classifier_activation="softmax", pretrained=None, **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.4, 1.8, fix_head_stem=True)
    first_conv_filter, output_conv_filter = kwargs.pop("first_conv_filter", first_conv_filter), kwargs.pop("output_conv_filter", output_conv_filter)
    return EfficientNetV1(**locals(), se_ratios=[0] * len(depthes), is_fused=False, model_name="efficientnet_v1-lite4", **kwargs)



import copy
import math
import torch
import numpy as np
from tqdm import tqdm
from pathlib import Path
from torch import nn
from torch.cuda import amp
from torch.optim import lr_scheduler

# from ultralytics.yolo.utils.torch_utils import ModelEMA


class FakeArgs:
    def __init__(self, **kwargs):
        self.update(**kwargs)

    def update(self, **kwargs):
        for kk, vv in kwargs.items():
            setattr(self, kk, vv)


class ModelEMA:
    """Updated Exponential Moving Average (EMA) from https://github.com/rwightman/pytorch-image-models
    Keeps a moving average of everything in the model state_dict (parameters and buffers)
    For EMA details see https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    To disable EMA set the `enabled` attribute to `False`.
    """

    def __init__(self, model, decay=0.9999, tau=2000, updates=0):
        # Create EMA
        self.ema = copy.deepcopy(model).eval()  # FP32 EMA
        self.updates = updates  # number of EMA updates
        self.decay = lambda x: decay * (1 - math.exp(-x / tau))  # decay exponential ramp (to help early epochs)
        for p in self.ema.parameters():
            p.requires_grad_(False)
        self.enabled = True

    def update(self, model):
        # Update EMA parameters
        if self.enabled:
            self.updates += 1
            d = self.decay(self.updates)

            msd = model.state_dict()  # model state_dict
            for k, v in self.ema.state_dict().items():
                if v.dtype.is_floating_point:  # true for FP16 and FP32
                    v *= d
                    v += (1 - d) * msd[k].detach()
                    # assert v.dtype == msd[k].dtype == torch.float32, f'{k}: EMA {v.dtype},  model {msd[k].dtype}'


def build_optimizer(model, name="sgd", lr=0.01, momentum=0.937, decay=5e-4):
    g = [], [], []  # optimizer parameter groups
    bn = tuple(v for k, v in nn.__dict__.items() if "Norm" in k)  # normalization layers, i.e. BatchNorm2d()
    for v in model.modules():
        if hasattr(v, "bias") and isinstance(v.bias, nn.Parameter):  # bias (no decay)
            g[2].append(v.bias)
        if isinstance(v, bn):  # weight (no decay)
            g[1].append(v.weight)
        elif hasattr(v, "weight") and isinstance(v.weight, nn.Parameter):  # weight (with decay)
            g[0].append(v.weight)

    name_lower = name.lower()
    if name_lower == "sgd":
        optimizer = torch.optim.SGD(g[2], lr=lr, momentum=momentum, nesterov=True)
    elif name_lower == "adamw":
        optimizer = torch.optim.AdamW(g[2], lr=lr, betas=(momentum, 0.999), weight_decay=0.0)
    optimizer.add_param_group({"params": g[0], "weight_decay": decay})  # add g0 with weight_decay
    optimizer.add_param_group({"params": g[1], "weight_decay": 0.0})  # add g1 (BatchNorm2d weights)
    return optimizer


def train(model, dataset_path="coco.yaml", batch_size=16, epochs=100, initial_epoch=0, optimizer_name="sgd", rect_val=False, overwrite_cfg=None):
    from keras_cv_attention_models.yolov8 import eval, data, losses

    if torch.cuda.is_available():
        model = model.cuda()
        use_amp = True
    else:
        model = model.cpu()
        use_amp = False

    warmup_epochs = 3
    close_mosaic = 10

    imgsz = (model.input_shape[2] or 640) if hasattr(model, "input_shape") else 640
    print(">>>> imgsz:", imgsz)
    cfg = FakeArgs(data=dataset_path, imgsz=imgsz, iou=0.7, single_cls=False, max_det=300, task="detect", mode="train", split="val", half=False, conf=None)
    cfg.update(augment=False, degrees=0.0, translate=0.1, scale=0.5, shear=0.0, perspective=0.0, hsv_h=0.015, hsv_s=0.7, hsv_v=0.4, flipud=0.0, fliplr=0.5)
    cfg.update(mask_ratio=4, overlap_mask=True, project=None, name=None, save_txt=False, save_hybrid=False, save_json=False, plots=False, verbose=True, seed=0)
    if overwrite_cfg is not None:
        cfg.update(**overwrite_cfg)
    # from ultralytics.yolo.cfg import get_cfg
    # from ultralytics.yolo.utils import DEFAULT_CFG
    # cfg = get_cfg(DEFAULT_CFG)
    # cfg.data = dataset_path

    train_loader, val_loader = data.get_data_loader(dataset_path=dataset_path, batch_size=batch_size, imgsz=imgsz, rect_val=rect_val)
    device = next(model.parameters()).device  # get model device
    num_classes = getattr(model, "num_classes", model.output_shape[-1] - 64)
    print(">>>> num_classes =", num_classes)
    compute_loss = losses.Loss(device=device, nc=num_classes)
    optimizer = build_optimizer(model, name=optimizer_name)
    ema = ModelEMA(model)
    # lf = lambda x: (x * (1 - 0.01) / warmup_epochs + 0.01) if x < warmup_epochs else ((1 - x / epochs) * (1.0 - 0.01) + 0.01)  # linear
    lf = lambda x: (1 - x / epochs) * (1.0 - 0.01) + 0.01  # linear
    scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lf)
    scaler = amp.GradScaler(enabled=use_amp)
    validator = eval.Validator(val_loader=val_loader, model=ema.ema, cfg=cfg)
    # validator = v8.detect.DetectionValidator(val_loader, save_dir=Path("./test"), args=copy.copy(cfg))

    nb = len(train_loader)
    nw = max(round(warmup_epochs * nb), 100)
    nbs = 64
    accumulate = max(round(nbs / batch_size), 1)
    warmup_bias_lr = 0.1
    momentum = 0.937
    warmup_momentum = 0.8
    last_opt_step = -1
    for epoch in range(initial_epoch, epochs):
        # self.run_callbacks('on_train_epoch_start')
        model.train()
        # Update attributes (optional)
        if epoch == (epochs - close_mosaic):
            print("Closing dataloader mosaic")
            if hasattr(train_loader.dataset, "mosaic"):
                train_loader.dataset.mosaic = False
            if hasattr(train_loader.dataset, "close_mosaic"):
                train_loader.dataset.close_mosaic(hyp=cfg)

        tloss = None
        optimizer.zero_grad()
        loss_names = ["box_loss", "cls_loss", "dfl_loss"]
        print(("\n" + "%11s" * (3 + len(loss_names))) % ("Epoch", *loss_names, "Instances", "Size"))
        pbar = tqdm(enumerate(train_loader), total=nb, bar_format="{l_bar}{bar:10}{r_bar}")
        for i, batch in pbar:
            # self.run_callbacks('on_train_batch_start')
            ni = i + nb * epoch
            if ni <= nw:
                xi = [0, nw]  # x interp
                accumulate = max(1, np.interp(ni, xi, [1, nbs / batch_size]).round())
                for j, x in enumerate(optimizer.param_groups):
                    # bias lr falls from 0.1 to lr0, all other lrs rise from 0.0 to lr0
                    x["lr"] = np.interp(ni, xi, [warmup_bias_lr if j == 0 else 0.0, x["initial_lr"] * lf(epoch)])
                    if "momentum" in x:
                        x["momentum"] = np.interp(ni, xi, [warmup_momentum, momentum])

            # Forward
            with torch.cuda.amp.autocast(use_amp):
                preds = model(batch["img"].to(device, non_blocking=True).float() / 255)
                loss, loss_items = compute_loss(preds, batch)
                tloss = (tloss * i + loss_items) / (i + 1) if tloss is not None else loss_items

            # Backward
            scaler.scale(loss).backward()

            # Optimize - https://pytorch.org/docs/master/notes/amp_examples.html
            if ni - last_opt_step >= accumulate:
                # optimizer_step(model, optimizer, scaler)
                scaler.unscale_(optimizer)  # unscale gradients
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=10.0)  # clip gradients
                scaler.step(optimizer)
                scaler.update()
                optimizer.zero_grad()
                ema.update(model)
                last_opt_step = ni

            loss_len = tloss.shape[0] if len(tloss.size()) else 1
            losses = tloss if loss_len > 1 else torch.unsqueeze(tloss, 0)
            pbar.set_description(("%11s" * 1 + "%11.4g" * (2 + loss_len)) % (f"{epoch + 1}/{epochs}", *losses, batch["cls"].shape[0], batch["img"].shape[-1]))
        scheduler.step()
        validator()

        if hasattr(model, "model") and hasattr(model.model, "save_weights"):
            model.model.save_weights(model.model.name + ".h5")
        elif hasattr(model, "model") and hasattr(model.model, "state_dict"):
            torch.save(model.model.state_dict(), model.__class__.__name__ + ".pth")

    # Save ema model
    ema_model = ema.ema.model
    if hasattr(ema_model, "model") and hasattr(ema_model.model, "save_weights"):
        ema_model.model.save_weights(ema_model.model.name + "_ema.h5")
    elif hasattr(ema_model, "model") and hasattr(ema_model.model, "state_dict"):
        torch.save(ema_model.model.state_dict(), ema_model.__class__.__name__ + "_ema.pth")
    return ema


if __name__ == "__main__":
    os.environ["KECAM_BACKEND"] = "torch"
    sys.path.append("../ultralytics/")
    from keras_cv_attention_models.yolov8 import train, yolov8, torch_wrapper

    # from ultralytics import YOLO
    # model = YOLO('../ultralytics/ultralytics/models/v8/yolov8n.yaml').model
    model = yolov8.YOLOV8_N(input_shape=(3, None, None), classifier_activation=None, pretrained=None)
    model = torch_wrapper.Detect(model)
    train.train(model, dataset_path="coco128.yaml")



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import activation_by_name, drop_block, eca_module, se_module, make_divisible, add_pre_post_process

PRETRAINED_DICT = {
    "nfnetf0": {"imagenet": "7f8ee8639d468597de41566ce1b481c7"},
    "nfnetf1": {"imagenet": "f5d298e50996f0a11a8b097e0f890fa2"},
    "nfnetf2": {"imagenet": "3b0f5d6ac33a2833d7d9ee0e02aae4bc"},
    "nfnetf3": {"imagenet": "e28864d7a553cdab9766223994e0a96d"},
    "nfnetf4": {"imagenet": "9a44cb37155f67b88b3900b7c2c9617d"},
    "nfnetf5": {"imagenet": "728d9202661de4d1003c9b149c25461e"},
    "nfnetf6": {"imagenet": "ee4a06b4a543531d72ea5a8a101336ac"},
    "nfnetl0": {"imagenet": "6bd4d11037bf720506aa3d3e12ec4f53"},
    "eca_nfnetl0": {"imagenet": "7789af74226ffee28a0a68cdca6f3737"},
    "eca_nfnetl1": {"imagenet": "cd17a98175825258d32229bc82b744fd"},
    "eca_nfnetl2": {"imagenet": "ca7e0bba4f2d1945d881ffc6e36bed36"},
}


NON_LINEAR_GAMMA = dict(
    identity=1.0,
    celu=1.270926833152771,
    elu=1.2716004848480225,
    gelu=1.7015043497085571,
    leaky_relu=1.70590341091156,
    log_sigmoid=1.9193484783172607,
    log_softmax=1.0002083778381348,
    relu=1.7139588594436646,
    relu6=1.7131484746932983,
    selu=1.0008515119552612,
    sigmoid=4.803835391998291,
    swish=1.7881293296813965,  # silu
    softsign=2.338853120803833,
    softplus=1.9203323125839233,
    tanh=1.5939117670059204,
)


@backend.register_keras_serializable(package="nfnets")
class ScaledStandardizedConv2D(layers.Conv2D):
    """
    Copied from https://github.com/google-research/big_transfer/blob/master/bit_tf2/models.py, Author: Lucas Beyer
    Modified reference: https://github.com/deepmind/deepmind-research/blob/master/nfnets/base.py#L121
    """

    def __init__(self, gamma=1.0, eps=1e-5, *args, **kwargs):
        super(ScaledStandardizedConv2D, self).__init__(*args, **kwargs)
        self.eps, self.gamma = eps, gamma

    def build(self, input_shape):
        super(ScaledStandardizedConv2D, self).build(input_shape)
        # Wrap a standardization around the conv OP.
        if hasattr(self, "_convolution_op"):
            default_conv_op = self._convolution_op  # TF < 2.7.0
        else:
            default_conv_op = self.convolution_op  # TF 2.7.0
        self.gain = self.add_weight(name="gain", shape=(self.filters,), initializer="ones", trainable=True)
        self.fan_in = float(np.prod(self.kernel.shape[:-1]))
        self.__eps__ = float(self.eps)
        self.__gamma__ = float(self.gamma)

        def standardized_conv_op(inputs, kernel):
            # Kernel has shape HWIO, normalize over HWI
            mean, var = functional.moments(kernel, axes=[0, 1, 2], keepdims=True)
            # Manually fused normalization, eq. to (w - mean) * gain / sqrt(N * var)
            # print(">>>>", mean.dtype, var.dtype, self.fan_in.dtype, self.__eps__.dtype, self.__gamma__.dtype, self.gain.dtype)
            scale = functional.rsqrt(functional.maximum(var * self.fan_in, self.__eps__)) * (self.gain * self.__gamma__)
            return default_conv_op(inputs, (kernel - mean) * scale)

        if hasattr(self, "_convolution_op"):
            self._convolution_op = standardized_conv_op  # TF < 2.7.0
        else:
            self.convolution_op = standardized_conv_op  # TF 2.7.0
        self.built = True

    def get_config(self):
        base_config = super(ScaledStandardizedConv2D, self).get_config()
        base_config.update({"eps": self.eps, "gamma": self.gamma})
        return base_config


@backend.register_keras_serializable(package="nfnets")
class ZeroInitGain(layers.Layer):
    def __init__(self, use_bias=False, weight_init_value=0, bias_init_value=0, **kwargs):
        super().__init__(**kwargs)
        self.use_bias = use_bias
        self.ww_init = initializers.Constant(weight_init_value) if weight_init_value != 0 else "zeros"
        self.bb_init = initializers.Constant(bias_init_value) if bias_init_value != 0 else "zeros"

    def build(self, input_shape):
        self.gain = self.add_weight(name="gain", shape=(), initializer=self.ww_init, trainable=True)
        if self.use_bias:
            self.bias = self.add_weight(name="bias", shape=(), initializer=self.bb_init, trainable=True)
        super().build(input_shape)

    def call(self, inputs):
        return (inputs * self.gain + self.bias) if self.use_bias else (inputs * self.gain)

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"use_bias": self.use_bias})
        return base_config


def std_conv2d_with_init(inputs, filters, kernel_size, strides=1, padding="valid", torch_padding=False, gamma=1.0, name=None, **kwargs):
    pad = max(kernel_size) // 2 if isinstance(kernel_size, (list, tuple)) else kernel_size // 2
    if torch_padding and padding.lower() == "same" and pad != 0:
        inputs = layers.ZeroPadding2D(padding=pad, name=name and name + "pad")(inputs)
        padding = "valid"

    return ScaledStandardizedConv2D(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        gamma=gamma,
        name=name and name + "conv",
        **kwargs,
    )(inputs)


def activation_by_name_with_gamma(inputs, activation="gelu", gamma=1.0, name=None):
    nn = activation_by_name(inputs, activation=activation, name=name)
    return nn if gamma == 1.0 else (nn * gamma)


def block(
    inputs,
    filters,
    beta=1.0,
    strides=1,
    drop_rate=0,
    alpha=0.2,
    channel_ratio=0.5,
    se_ratio=0.5,
    group_size=128,
    use_zero_init_gain=True,
    torch_padding=False,
    attn_type="se",
    conv_gamma=1.0,
    act_gamma=1.0,
    activation="gelu",
    name="",
):
    hidden_filter = int(filters * channel_ratio)
    attn_gain = 2.0
    # print(f">>>> {beta = }")
    preact = activation_by_name_with_gamma(inputs, activation, gamma=act_gamma, name=name + "preact_") * beta

    if strides > 1 or inputs.shape[-1] != filters:
        if strides > 1:
            shortcut = layers.AvgPool2D(strides, strides=strides, padding="same", name=name + "shorcut_down")(preact)
        else:
            shortcut = preact
        shortcut = std_conv2d_with_init(shortcut, filters, 1, strides=1, gamma=conv_gamma, name=name + "shortcut_")
    else:
        shortcut = inputs

    groups = hidden_filter // group_size
    conv_params_3 = {"kernel_size": 3, "padding": "same", "torch_padding": torch_padding, "gamma": conv_gamma}
    deep = std_conv2d_with_init(preact, hidden_filter, 1, strides=1, gamma=conv_gamma, name=name + "deep_1_")
    deep = activation_by_name_with_gamma(deep, activation, gamma=act_gamma, name=name + "deep_1_")
    deep = std_conv2d_with_init(deep, hidden_filter, strides=strides, **conv_params_3, groups=groups, name=name + "deep_2_")
    deep = activation_by_name_with_gamma(deep, activation, gamma=act_gamma, name=name + "deep_2_")
    deep = std_conv2d_with_init(deep, hidden_filter, strides=1, **conv_params_3, groups=groups, name=name + "deep_3_")  # Extra conv
    deep = activation_by_name_with_gamma(deep, activation, gamma=act_gamma, name=name + "deep_3_")
    deep = std_conv2d_with_init(deep, filters, 1, strides=1, gamma=conv_gamma, name=name + "deep_4_")
    if se_ratio > 0 and attn_type == "se":
        deep = se_module(deep, se_ratio=se_ratio, activation="relu", use_bias=True, name=name + "se_")
        deep *= attn_gain
    elif attn_type == "eca":
        deep = eca_module(deep, name=name + "eca_")
        deep *= attn_gain

    deep = drop_block(deep, drop_rate)
    if use_zero_init_gain:
        deep = ZeroInitGain(name=name + "deep_gain")(deep)
    deep *= alpha
    return layers.Add(name=name + "output")([shortcut, deep])


def stack(inputs, blocks, filters, betas=1.0, strides=2, stack_drop=0, block_params={}, name=""):
    nn = inputs
    for id in range(blocks):
        cur_strides = strides if id == 0 else 1
        block_name = name + "block{}_".format(id + 1)
        block_drop_rate = stack_drop[id] if isinstance(stack_drop, (list, tuple)) else stack_drop
        beta = betas[id] if isinstance(stack_drop, (list, tuple)) else betas
        nn = block(nn, filters, beta, cur_strides, block_drop_rate, name=block_name, **block_params)
    return nn


def stem(inputs, stem_width, activation="gelu", torch_padding=False, conv_gamma=1.0, act_gamma=1.0, name=""):
    conv_params = {"kernel_size": 3, "padding": "same", "torch_padding": torch_padding, "gamma": conv_gamma}
    nn = std_conv2d_with_init(inputs, stem_width // 8, strides=2, **conv_params, name=name + "1_")
    nn = activation_by_name_with_gamma(nn, activation, gamma=act_gamma, name=name + "1_")
    nn = std_conv2d_with_init(nn, stem_width // 4, strides=1, **conv_params, name=name + "2_")
    nn = activation_by_name_with_gamma(nn, activation, gamma=act_gamma, name=name + "2_")
    nn = std_conv2d_with_init(nn, stem_width // 2, strides=1, **conv_params, name=name + "3_")
    nn = activation_by_name_with_gamma(nn, activation, gamma=act_gamma, name=name + "3_")
    nn = std_conv2d_with_init(nn, stem_width, strides=2, **conv_params, name=name + "4_")
    return nn


def NormFreeNet(
    num_blocks,
    attn_type="se",
    stem_width=128,
    out_channels=[256, 512, 1536, 1536],
    channel_ratio=0.5,
    num_features_factor=2,
    strides=[1, 2, 2, 2],
    input_shape=(224, 224, 3),
    num_classes=1000,
    se_ratio=0.5,
    group_size=128,
    use_zero_init_gain=True,
    torch_padding=False,
    gamma_in_act=True,
    alpha=0.2,
    width_factor=1.0,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained="imagenet",
    model_name="nfnet",
    kwargs=None,
):
    if gamma_in_act:
        # activation.split("/")[0] for supporting `gelu/app`
        conv_gamma, act_gamma = 1.0, NON_LINEAR_GAMMA.get(activation.split("/")[0], 1.0)
    else:
        act_gamma, conv_gamma = 1.0, NON_LINEAR_GAMMA.get(activation.split("/")[0], 1.0)

    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(shape=input_shape)
    stem_width = make_divisible(stem_width * width_factor, 8)
    nn = stem(inputs, stem_width, activation=activation, torch_padding=torch_padding, conv_gamma=conv_gamma, act_gamma=act_gamma, name="stem_")

    block_params = {  # params same for all blocks
        "alpha": alpha,
        "channel_ratio": channel_ratio,
        "se_ratio": se_ratio,
        "group_size": group_size,
        "use_zero_init_gain": use_zero_init_gain,
        "torch_padding": torch_padding,
        "attn_type": attn_type,
        "conv_gamma": conv_gamma,
        "act_gamma": act_gamma,
        "activation": activation,
    }

    drop_connect_rates = functional.split(functional.linspace(0.0, drop_connect_rate, sum(num_blocks)), num_blocks)
    drop_connect_rates = [ii.numpy().tolist() for ii in drop_connect_rates]
    beta_list = [(1 + alpha**2 * ii) ** -0.5 for ii in range(max(num_blocks) + 1)]
    pre_beta = 1.0
    for id, (num_block, out_channel, stride, drop_connect) in enumerate(zip(num_blocks, out_channels, strides, drop_connect_rates)):
        name = "stack{}_".format(id + 1)
        out_channel = make_divisible(out_channel * width_factor, 8)
        betas = beta_list[: num_block + 1]
        betas[0] = pre_beta
        nn = stack(nn, num_block, out_channel, betas, stride, drop_connect, block_params, name=name)
        pre_beta = betas[-1]

    if num_features_factor > 0:
        output_conv_filter = make_divisible(num_features_factor * out_channels[-1] * width_factor, 8)
        nn = std_conv2d_with_init(nn, output_conv_filter, 1, gamma=conv_gamma, name="post_")
    nn = activation_by_name_with_gamma(nn, activation, gamma=act_gamma, name="post_")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="nfnets", pretrained=pretrained)
    return model


@register_model
def NFNetF0(input_shape=(256, 256, 3), num_classes=1000, activation="gelu", dropout=0.2, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[1, 2, 6, 3], model_name="nfnetf0", **locals(), **kwargs)


@register_model
def NFNetF1(input_shape=(320, 320, 3), num_classes=1000, activation="gelu", dropout=0.3, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[2, 4, 12, 6], model_name="nfnetf1", **locals(), **kwargs)


@register_model
def NFNetF2(input_shape=(352, 352, 3), num_classes=1000, activation="gelu", dropout=0.4, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[3, 6, 18, 9], model_name="nfnetf2", **locals(), **kwargs)


@register_model
def NFNetF3(input_shape=(416, 416, 3), num_classes=1000, activation="gelu", dropout=0.4, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[4, 8, 24, 12], model_name="nfnetf3", **locals(), **kwargs)


@register_model
def NFNetF4(input_shape=(512, 512, 3), num_classes=1000, activation="gelu", dropout=0.5, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[5, 10, 30, 15], model_name="nfnetf4", **locals(), **kwargs)


@register_model
def NFNetF5(input_shape=(544, 544, 3), num_classes=1000, activation="gelu", dropout=0.5, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[6, 12, 36, 18], model_name="nfnetf5", **locals(), **kwargs)


@register_model
def NFNetF6(input_shape=(576, 576, 3), num_classes=1000, activation="gelu", dropout=0.5, pretrained="imagenet", **kwargs):
    return NormFreeNet(num_blocks=[7, 14, 42, 21], model_name="nfnetf6", **locals(), **kwargs)


@register_model
def NFNetF7(input_shape=(608, 608, 3), num_classes=1000, activation="gelu", dropout=0.5, pretrained=None, **kwargs):
    return NormFreeNet(num_blocks=[8, 16, 48, 24], model_name="nfnetf7", **locals(), **kwargs)


def NormFreeNet_Light(channel_ratio=0.25, group_size=64, torch_padding=True, use_zero_init_gain=False, gamma_in_act=False, **kwargs):
    kwargs.pop("kwargs", None)
    return NormFreeNet(**locals(), **kwargs)


@register_model
def NFNetL0(input_shape=(288, 288, 3), num_classes=1000, activation="swish", dropout=0.2, pretrained="imagenet", **kwargs):
    num_blocks = [1, 2, 6, 3]
    num_features_factor = 1.5
    se_ratio = 0.25
    return NormFreeNet_Light(model_name="nfnetl0", **locals(), **kwargs)


@register_model
def ECA_NFNetL0(input_shape=(288, 288, 3), num_classes=1000, activation="swish", dropout=0.2, pretrained="imagenet", **kwargs):
    num_blocks = [1, 2, 6, 3]
    num_features_factor = 1.5
    attn_type = "eca"
    return NormFreeNet_Light(model_name="eca_nfnetl0", **locals(), **kwargs)


@register_model
def ECA_NFNetL1(input_shape=(320, 320, 3), num_classes=1000, activation="swish", dropout=0.2, pretrained="imagenet", **kwargs):
    num_blocks = [2, 4, 12, 6]
    attn_type = "eca"
    return NormFreeNet_Light(model_name="eca_nfnetl1", **locals(), **kwargs)


@register_model
def ECA_NFNetL2(input_shape=(384, 384, 3), num_classes=1000, activation="swish", dropout=0.2, pretrained="imagenet", **kwargs):
    num_blocks = [3, 6, 18, 9]
    attn_type = "eca"
    return NormFreeNet_Light(model_name="eca_nfnetl2", **locals(), **kwargs)


@register_model
def ECA_NFNetL3(input_shape=(448, 448, 3), num_classes=1000, activation="swish", dropout=0.2, pretrained=None, **kwargs):
    num_blocks = [4, 8, 24, 12]
    attn_type = "eca"
    return NormFreeNet_Light(model_name="eca_nfnetl3", **locals(), **kwargs)
